# Copyright (c) 2010-2017 Renaud METRICH
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products
#    derived from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package AdvPowerMgr::Class;

use strict;
use warnings;

use base qw( Exporter );

use Scalar::Util qw( blessed );
use Symbol qw( qualify_to_ref );

__PACKAGE__->_mk_accessors(qw( version disable_normalizers ));

require AdvPowerMgr::Errors;

#
# Constructor parameters ($params):
# - disable_normalizers (1|0): disable normalizers completely, can be reactivated
# - delay_triggers (1|0):  don't run triggers, but enqueue them
sub new {
    my $class = shift;
    my $params = defined($_[0]) ? { %{$_[0]} } : {}; shift;

    $params->{version} = 1 unless defined($params->{version});
    $params->{disable_normalizers} = 0 unless defined($params->{disable_normalizers});

    # If instructed to delay triggers, start at 1
    $params->{delayed_depth} = defined($params->{delay_triggers}) && $params->{delay_triggers} ? 1 : 0;

    my $self = {
        delayed_triggers => [],
        delayed_depth => $params->{delayed_depth}
    };
    bless($self, $class);
    $self->disable_normalizers($params->{disable_normalizers}, { disable_trigger => 1 });
    $self->version($params->{version});

    return $self;
}

sub delay_triggers {
    my $self = shift;

    unless (blessed($self)) {
        AdvPowerMgr::Errors->param_blessed($self);
    }
    $self->{delayed_depth}++;
}

sub run_delayed_triggers {
    my $self = shift;

    unless (blessed($self)) {
        AdvPowerMgr::Errors->param_blessed($self);
    }
    if ($self->{delayed_depth} == 0) {
        my (undef, $filename, $line) = caller(1);
        AdvPowerMgr::Errors->die_ie("depth < 0", $filename, $line);
    }
    $self->{delayed_depth}--;
    if ($self->{delayed_depth} == 0 && @{$self->{delayed_triggers}}) {
        no strict 'refs';
        while (my $e = pop(@{$self->{delayed_triggers}})) {
            my $func = $e->[0];
            $self->$func($e->[1]);
        }
    }
}

sub disable_normalizers_validator {
    my $class = shift;
    my $value = shift;
    require AdvPowerMgr::Validators;
    return AdvPowerMgr::Validators->boolean_validator($value);
}

sub _mk_accessors {
    my $class = shift;

    no strict 'refs';
    foreach my $field (@_) {
        my $ref = qualify_to_ref($field, $class);

        *{$ref} = sub {
            my $self = shift;
            if (@_) {
                # Setter options
                # - disable_validator
                # - disable_normalizer
                # - disable_trigger
                my $val = shift;
                my $options = shift || {};
                my $func = "${field}_validator";
                if ($self->can($func) && !$options->{disable_validator}) {
                    $val = $self->$func($val);
                }
                $func = "${field}_normalizer";
                if ($self->can($func) && !$options->{disable_normalizer} && !$self->disable_normalizers()) {
                    $val = $self->$func($val);
                }
                my $was = $self->{$field};
                $self->{$field} = $val;
                $func = "${field}_trigger";
                if ($self->can($func) && !$options->{disable_trigger}) {
                    # When trigger is called on __PACKAGE__, cannot delay it
                    if (!blessed($self) || $self->{delayed_depth} == 0) {
                        $self->$func($was);
                    } else {
                        push(@{$self->{delayed_triggers}}, [ $func, $was ]);
                    }
                }
                return $was;
            } else {
                # Getter
                return $self->{$field};
            }
        };
    }
}

#
# Die on Warn when unit testing
#
if (defined($ENV{UNIT_TESTING})) {
    $SIG{__WARN__} = sub {
        CORE::die "@_";
    };
}

1;

# vi:set autoindent tabstop=8 shiftwidth=4 softtabstop=4 expandtab:
