# Copyright (c) 2017 Renaud METRICH
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products
#    derived from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package AdvPowerMgr::Date;

use strict;
use warnings;

use Scalar::Util qw( blessed );
use Time::Local qw( timelocal );
use overload
#    fallback => 1,
    '""'    => "stringify",
    '0+'    => "numify",
    '<=>'   => "numcompare",
    "+"     => "increment",
    "-"     => "decrement";

use AdvPowerMgr::Errors;

use base qw( AdvPowerMgr::Class );

__PACKAGE__->_mk_accessors(qw( day month year ts _weekday _yearday _weeknum ));

#
# An object conforming to ISO 8601 (see https://en.wikipedia.org/wiki/ISO_8601):
# - First day in week is 'Mon', last day is 'Sun'
# - 'Mon' day number is 1, 'Sun' day number is 7
# - First week in year contains the first thursday
#

our @weekday = ( "dummy", "mon", "tue", "wed", "thr", "fri", "sat", "sun" );

sub day_validator {
    my $self = shift;
    my $value = shift;

    eval {
        timelocal(0, 0, 0, $value, $self->month() - 1, $self->year() - 1900);
    };
    die_ie(sprintf("Unexpected 'day' value '%d': %s", $value, $@)) if $@;

    return $value;
}

sub day_trigger {
    my $self = shift;

    $self->_compute_ts_and_rest();
}

sub month_validator {
    my $self = shift;
    my $value = shift;

    eval {
        timelocal(0, 0, 0, $self->day(), $value - 1, $self->year() - 1900);
    };
    die_ie(sprintf("Unexpected 'month' value '%d': %s", $value, $@)) if $@;

    return $value;
}

sub month_trigger {
    my $self = shift;

    $self->_compute_ts_and_rest();
}

sub year_validator {
    my $self = shift;
    my $value = shift;

    eval {
        timelocal(0, 0, 0, $self->day(), $self->month() - 1, $value - 1900);
    };
    die_ie(sprintf("Unexpected 'year' value '%d': %s", $value, $@)) if $@;

    return $value;
}

sub year_trigger {
    my $self = shift;

    $self->_compute_ts_and_rest();
}

sub ts_validator {
    my $self = shift;
    my $value = shift;

    die_ie(sprintf("Unexpected 'ts' value '%d'", $value)) unless $value >= 0;

    return $value;
}

sub ts_normalizer {
    my $self = shift;
    my $value = shift;

    # Always return the value at midnight
    my (undef, undef, undef, $dm, $mo, $yo, $wd, $yd) = localtime($value);
    return timelocal(0, 0, 0, $dm, $mo, $yo);
}

sub ts_trigger {
    my $self = shift;

    my (undef, undef, undef, $dm, $mo, $yo, $wd, $yd) = localtime($self->ts());
    $self->day($dm, { disable_validator => 1, disable_trigger => 1 });
    $self->month($mo + 1, { disable_validator => 1, disable_trigger => 1 });
    $self->year($yo + 1900, { disable_validator => 1, disable_trigger => 1 });
    $self->_weekday($wd == 0 ? 7 : $wd, { disable_validator => 1, disable_trigger => 1 });
    $self->_yearday($yd + 1, { disable_validator => 1, disable_trigger => 1 });
    $self->_weeknum(undef); # Compute on demand
}

sub _compute_ts_and_rest {
    my $self = shift;

    my $ts = timelocal(0, 0, 0, $self->day(), $self->month() - 1, $self->year() - 1900);
    $self->ts($ts, { disable_validator => 1, disable_normalizer => 1, disable_trigger => 1 });
    my (undef, undef, undef, undef, undef, undef, $wd, $yd) = localtime($ts);
    $self->_weekday($wd == 0 ? 7 : $wd, { disable_validator => 1, disable_trigger => 1 });
    $self->_yearday($yd + 1, { disable_validator => 1, disable_trigger => 1 });
    $self->_weeknum(undef); # Compute on demand
}

sub new {
    my $class = shift;
    my $when = param_defined(shift);

    # when can be:
    # - a AdvPowerMgr::Date => makes a copy
    # - a HASH ref { year, month, day }
    # - a timestamp in local timezone

    my $params = {};
    my $self = $class->SUPER::new($params);
    bless($self, $class);

    if (blessed($when)) {
        die_ie(sprintf("Unexpected class '%s'", ref($when))) unless $when->isa(__PACKAGE__);
        $self->ts($when->ts(), { disable_validator => 1 });
        $self->_weeknum($when->_weeknum());
    } else {
        if (ref($when) eq "HASH") {
            eval {
                $when = timelocal(0, 0, 0, $when->{day}, $when->{month} - 1, $when->{year} - 1900);
            };
            die_ie(sprintf("Unexpected HASH 'when': %s", $@)) if $@;
        }
        $self->ts($when);
    }

    return $self;
}

sub weekday {
    my $self = shift;
    return $self->_weekday();
}

sub yearday {
    my $self = shift;
    return $self->_yearday();
}

#
# Return the week number
#
sub weeknum {
    my $self = shift;

    my $weeknum = $self->_weeknum();

    unless (defined($weeknum)) {
        # Jan 4 is ALWAYS in Week 1
        my $yearday = $self->yearday();
        my $Jan4 = AdvPowerMgr::Date->new({ year => $self->year(), month => 1, day => 4 });
        my $index = $Jan4->weekday();
        my $first_week_yearday = $Jan4->yearday() - $index + 1;

        if ($self->yearday() < $first_week_yearday) {
            $Jan4 = AdvPowerMgr::Date->new({ year => $self->year() - 1, month => 1, day => 4 });
            $index = $Jan4->weekday();
            $first_week_yearday = $Jan4->yearday() - $index + 1;

            $yearday += AdvPowerMgr::Date->new({ year => $self->year() - 1, month => 12, day => 31 })->yearday();
        }

        $weeknum = int(($yearday - $first_week_yearday) / 7) + 1;
        if ($weeknum == 53) {
            $Jan4 = AdvPowerMgr::Date->new({ year => $self->year() + 1, month => 1, day => 4 });
            $index = $Jan4->weekday();
            $first_week_yearday = $Jan4->yearday() - $index + 1;
            if ($self->yearday() >= AdvPowerMgr::Date->new({ year => $self->year(), month => 12, day => 31 })->yearday() + $first_week_yearday) {
                $weeknum = 1;
            }
        }
        $self->_weeknum($weeknum);
    }
    return $weeknum;
}

#
# Increment by a number of days
#
sub increment {
    my ($self, $other, $swap, $is_decr) = @_; # is_decr: additional flag to indicate decrement) 
    my $ndays;

    die_ie(sprintf("Unexpected argument for 'swap': '%s'", $swap)) if (defined($swap) && $swap ne "");
    if (!ref($other)) {
        $ndays = $other;
    } else {
        die_ie(sprintf("Unexpected argument '%s'", $other));
    }

    my $newobj = $self;
    if (defined($swap)) {
        $newobj = AdvPowerMgr::Date->new($self);
    }

    # Compute based on Date @ 12:00 (no timezone changes daylight saving at that time)
    my $ts = timelocal(0, 0, 12, $newobj->day(), $newobj->month() - 1, $newobj->year() - 1900);
    if (defined($is_decr)) {
        $ts -= $ndays * 86400;
    } else {
        $ts += $ndays * 86400;
    }
    $newobj->ts($ts);

    return $newobj;
}

#
# Decrement by a number of days
#
sub decrement {
    my ($self, $other, $swap) = @_;
    return $self->increment($other, $swap, 1);
}

sub numify {
    my $self = shift;
    return $self->ts();
}

sub numcompare {
    my ($self, $other) = @_;

    return $self->ts() <=> $other->ts();
}

sub stringify {
    my $self = shift;
    return sprintf("%04d/%02d/%02d (%s, ts=%d)",
        $self->year(), $self->month(), $self->day(),
        $weekday[$self->weekday()], $self->ts());
}

1;

# vi:set autoindent tabstop=8 shiftwidth=4 softtabstop=4 expandtab:
