# Copyright (c) 2017 Renaud METRICH
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products
#    derived from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package AdvPowerMgr::Plugin::GenericText;

use strict;
use warnings;

use base qw( AdvPowerMgr::Class );

use AdvPowerMgr::Errors;
use AdvPowerMgr::I18N qw( i18n );
use AdvPowerMgr::Log;
use AdvPowerMgr::Utils qw( range list_splitter range_splitter );

__PACKAGE__->_mk_accessors(qw( text _repr ));

use overload '""' => "stringify";

sub new {
    my $class = shift;
    my $text = param_defined(shift);;
    my $params = shift; defined($params) or $params = {};

    # Do not normalize by default, no need for that
    $params->{disable_normalizers} = 1 unless defined($params->{disable_normalizers});

    my $self = $class->SUPER::new($params);
    bless($self, $class);

    $self->text($text);

    return $self;
}

sub stringify {
    my $self = shift;
    my @class = split(/::/, ref($self));
    return sprintf("%s('%s')", $class[-1], $self->text());
}

sub disable_normalizers_trigger {
    my $self = shift;
    my $was = shift;

    return if $was == 0;
    # Normalize text, but don't recompute internal representation (unneeded)
    $self->text($self->text_normalizer($self->text()), { disable_trigger => 1 });
}

sub _text_validator_i18n_msg {
    return 'plugin.unexpected_string';
}

sub text_validator {
    my $self = shift;
    my $value = param_defined(shift);

    my $regex = "^" . $self->_regex() . "\$";
    die i18n($self->_text_validator_i18n_msg(), $value) unless $value =~ m#$regex#i;

    return $value;
}

sub text_normalizer {
    my $class = shift;
    my $value = param_defined(shift);

    $value =~ s#\s+# #g;
    $value =~ s#\s*-\s*#-#g;
    $value =~ s#\s*,\s*#,#g;

    return $value;
}

sub text_trigger {
    my $self = shift;

    $self->_repr([ $self->_repr_map($self->text()) ]);
}

# Static method mapping an item to internal representation
# Defaults to noop
sub _repr_map {
    my $class = shift;
    my $item = param_defined(shift);

    return $item;
}

# Static method mapping a range of items to internal representation
sub _repr_map_range {
    my $class = shift;
    my $first = shift;
    my $last = shift;

    die_ie("ABSTRACT");
}

sub _regex {
    my $class = shift;

    die_ie("ABSTRACT");
}

# Split list and map to internal representation
sub list_split_and_map {
    my $self = shift;
    my $baseregex = param_defined(shift);

    my @res = list_splitter($baseregex, $self->text());
    @res = map { $self->_repr_map($_) } @res;

    return @res;
}

# Split list of ranges and map to internal representation
sub lrange_split_and_map {
    my $self = shift;
    my $baseregex = param_defined(shift);

    my @res = list_splitter(range($baseregex), $self->text());
    @res = map { [ range_splitter($baseregex, $_) ] } @res;
    @res = map { $self->_repr_map_range($_->[0], $_->[1]) } @res;

    return @res;
}

1;

# vi:set autoindent tabstop=8 shiftwidth=4 softtabstop=4 expandtab:
