# Copyright (c) 2010-2017 Renaud METRICH
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products
#    derived from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package AdvPowerMgr::Plugin::PostShutdown::Connections;

use strict;
use warnings;

use base qw( AdvPowerMgr::Plugin::PostShutdown );

use JSON qw();

use AdvPowerMgr::Errors;
use AdvPowerMgr::I18N qw( i18n );
use AdvPowerMgr::Log;
use AdvPowerMgr::Plugin::PostShutdown::Connections::Line;
use AdvPowerMgr::System;
use AdvPowerMgr::Utils qw( ipv4_to_hex orregex );
use AdvPowerMgr::Validators qw( boolean_validator );

__PACKAGE__->_mk_accessors(qw( list lazy _repr ));

# COMMENTS := ( # | @ | - ) TEXT
our $comments_re = __PACKAGE__->comments_re();
our $line = AdvPowerMgr::Plugin::PostShutdown::Connections::Line->_regex();
our $re = qr/^\s*($line)\s*$/i;

sub list_validator {
    my $self = shift;
    my $value = shift;

    foreach my $item (@$value) {
        next if $item =~ $comments_re;
        die i18n('connections.invalid_line', $item) unless $item =~ $re;
    }

    return $value;
}

sub list_normalizer {
    my $self = shift;
    my $value = shift;

    my @res = ();

    foreach my $item (@$value) {
        if ($item =~ $comments_re) {
            push(@res, $item);
        } else {
            # uncoverable branch true
            $item =~ $re or die_ie(sprintf("%s: no match!", __PACKAGE__));
            push(@res, AdvPowerMgr::Plugin::PostShutdown::Connections::Line->text_normalizer($1));
        }
    }

    return \@res;
}

sub list_trigger {
    my $self = shift;

    my @res = ();

    foreach my $item (@{$self->list()}) {
        next if $item =~ $comments_re; # skip comments

        # uncoverable branch true
        $item =~ $re or die_ie(sprintf("%s: no match!", __PACKAGE__));
        push(@res, AdvPowerMgr::Plugin::PostShutdown::Connections::Line->new($1));
    }

    unless (@res) {
        # Considered all connections and ports by default
        push(@res, AdvPowerMgr::Plugin::PostShutdown::Connections::Line->new("*"));
    }

    $self->_repr(\@res);
}

sub lazy_validator {
    my $self = shift;
    my $value = shift;

    return boolean_validator($value);
}

sub lazy_trigger {
    my $self = shift;
    my $was = shift;

    $self->_reset() if defined($was) && $was != $self->lazy();
}

sub new {
    my $class = shift;
    my $params = defined($_[0]) ? { %{$_[0]} } : {}; shift;

    debug_func_enter($params);

    my $count = 0;
    $count++ if defined($params->{list});
    $count++ if defined($params->{connections});
    $count++ if defined($params->{ports});
    unless ($count <= 1) {
        my (undef, $filename, $line) = caller(1);
        die_ie("'connections', 'ports' and 'list' are mutually exclusive", $filename, $line);
    }
    $params->{version} = 2 unless defined($params->{version});

    if ($params->{version} == 1) {
        # 'ports' was the old name, now 'connections'
        # 'lan_only' vanished and was the default
        if (defined($params->{ports})) {
            my $lan_only = defined($params->{lan_only}) ? $params->{lan_only} : 1;
            my @res = ();
            my $found = 0;
            foreach my $item (split(/\n/, $params->{ports})) {
                if ($item =~ $comments_re) {
                    push(@res, $item);
                    next;
                }
                if ($lan_only) {
                    push(@res, "lan  $item");
                } else {
                    push(@res, $item);
                }
                $found++;
            }
            if (!$found && $lan_only) {
                # Plugin version < 4 was considering all ports by default on local LAN
                push(@res, "") if @res;
                push(@res, "## Automatically added for backward compatibility ##");
                push(@res, "lan  LPORTS  *");
            }
            $params->{list} = [ @res ];
        }
        $params->{version} = 2;
    }

    $params->{disable_normalizers} = 0 unless defined($params->{disable_normalizers}); 
    $params->{list} = [] unless defined($params->{list});
    $params->{list} = [ split(/\n/, $params->{connections}) ] if defined($params->{connections});
    $params->{lazy} = 1 unless defined($params->{lazy});

    if (defined($params->{objtype}) && $params->{objtype} eq "AdvPowerMgr::Plugin::Connections") {
        # Upgrade from 3.x, x <= 2
        delete $params->{objtype};
    }

    my $delayed_triggers = defined($params->{delay_triggers}) && $params->{delay_triggers} ? 1 : 0;
    $params->{delay_triggers} = 1 unless $delayed_triggers;
    my $self = $class->SUPER::new($params);
    bless($self, $class);

    $self->delay_triggers();
    $self->list($params->{list});
    $self->lazy($params->{lazy});
    $self->run_delayed_triggers();
    $delayed_triggers or $self->run_delayed_triggers();

    return $self;
}

sub to_hash {
    my $self = shift;
    my $options = shift || {};

    my $ret = $self->SUPER::to_hash($options);
    $ret->{connections} = join("\n", @{$self->list()});
    if ($options->{for_json}) {
        $ret->{disable_normalizers} = $self->disable_normalizers() ? JSON->true() : JSON->false();
        $ret->{lazy} = $self->lazy() ? JSON->true() : JSON->false();
    } else {
        $ret->{disable_normalizers} = $self->disable_normalizers();
        $ret->{lazy} = $self->lazy();
    }

    return $ret;
}

sub _value_reader {
    my $self = shift;

    my $count = 0;
    my %hosts = ();

    my @conns = AdvPowerMgr::System->get_connections();
    if (@conns != 0) {
        foreach my $c (@conns) {
            my ($r_ip, $l_port) = ($c->{remote_ip}, $c->{local_port});
            my $r_ip_hex = ipv4_to_hex($r_ip);
            next if $self->lazy() && exists($hosts{$r_ip_hex});

            foreach my $item (@{$self->_repr()}) {
                next unless $item->is_a_match($r_ip, $l_port);
                # Falling here means we had a match and no matching except clause
                info("Host %s connected to local port %d", $r_ip, $l_port);
                unless (exists($hosts{$r_ip_hex})) {
                    $hosts{$r_ip_hex} = { remote_ip => $r_ip, local_ports => [] };
                }
                $count++;
                push(@{$hosts{$r_ip_hex}->{local_ports}}, int($l_port));
                last; # foreach my $item
            }
        }
    }
    info("There are %d connections", $count);
    my @hosts = ();
    foreach my $hex (sort { $a <=> $b } keys(%hosts)) {
        $hosts{$hex}->{local_ports} = [ sort { $a <=> $b } @{$hosts{$hex}->{local_ports}} ];
        push(@hosts, $hosts{$hex});
    }
    $self->_laststate()->{hosts} = [ @hosts ];
    $self->_laststate()->{count} = $count;

    return $count;
}

sub _reset {
    my $self = shift;

    $self->SUPER::_reset();
    # Reset internal representation to destroy any cache
    $self->list_trigger();

    if ($self->enabled()) {
        my $state = $self->_laststate();
        $state->{count} = 0;
        $state->{hosts} = [];
        $state->{lazy} = int($self->lazy());
    }
}

# compute() is expected to return a hash with the following fields:
# - enabled
#   - 0 if plugin is disabled
#   - 1 otherwise
# - nextupdate_ts
#   - timestamp
# - done
#   - 0 if condition is not met
#   - 1 otherwise
# - checks_per_interval (static)
# - interval_min (static)
# - lazy (static)
# - count
# - hosts

1;

# vi:set autoindent tabstop=8 shiftwidth=4 softtabstop=4 expandtab:
