# Copyright (c) 2017 Renaud METRICH
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products
#    derived from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# NET := ( 0 .. 32 )
# BYTE := ( 0 .. 255 )
# SINGLE_CIDR := BYTE.BYTE.BYTE.BYTE[/NET]
package AdvPowerMgr::Plugin::PostShutdown::Connections::SingleCIDR;

use strict;
use warnings;

use base qw( AdvPowerMgr::Plugin::GenericText );

use AdvPowerMgr::Errors;
use AdvPowerMgr::Utils qw( ipv4_to_hex );

sub _text_validator_i18n_msg {
    return 'connections.unexpected_cidr_string';
}

our $byte = "(?:[0-1]?[0-9]?[0-9]|2[0-4][0-9]|25[0-5])";
our $ipv4 = join("\\.", ($byte) x 4);
our $net = "(?:[0-2]?[0-9]|3[0-2])";

our $regex = "$ipv4(?:/$net)?";
our $regex_matcher = "($ipv4)(?:/($net))?";

sub _regex {
    my $class = shift;

    return $regex;
}

sub _repr_map {
    my $class = shift;
    my $item = param_defined(shift);

    $item =~ m#^$regex_matcher$#i;
    my $hex = ipv4_to_hex($1);
    my $net = defined($2) ? int($2) : 32;
    $net = 0xFFFFFFFF ^ (0xFFFFFFFF >> $net);
    $hex &= $net;

    return [ $hex, $net ];
}

sub is_a_match {
    my $self = shift;
    my $value = param_defined(shift);

    unless ($value =~ m#^$ipv4$#i) {
        my (undef, $filename, $line) = caller(1);
        die_ie("unexpected value '$value'", $filename, $line);
    }
    my $hex = ipv4_to_hex($value);

    return scalar(grep { ($hex & $_->[1]) == $_->[0] } @{$self->_repr()}) ? 1 : 0;
}

sub text_normalizer {
    my $class = shift;
    my $value = param_defined(shift);

    $value =~ s#\b($byte)\b#int($1)#ieg;
    return $class->SUPER::text_normalizer($value);
}

# CIDR_LIST := LIST(SINGLE_CIDR)
package AdvPowerMgr::Plugin::PostShutdown::Connections::CIDRList;

use strict;
use warnings;

use base qw( AdvPowerMgr::Plugin::PostShutdown::Connections::SingleCIDR );

use AdvPowerMgr::Utils qw( list uniq );

sub _regex {
    my $class = shift;
    return list($class->SUPER::_regex());
}

sub text_trigger {
    my $self = shift;

    $self->_repr([ uniq(sort($self->list_split_and_map($self->SUPER::_regex()))) ]);
}

# CIDR := ( * | any | LAN | CIDR_LIST )
package AdvPowerMgr::Plugin::PostShutdown::Connections::CIDR;

use strict;
use warnings;

use base qw( AdvPowerMgr::Plugin::PostShutdown::Connections::CIDRList );

use AdvPowerMgr::Errors;
use AdvPowerMgr::System qw( );
use AdvPowerMgr::Utils qw( orregex );

sub _regex {
    my $class = shift;
    return orregex("\\*", "any", "LAN", $class->SUPER::_regex());
}

sub text_normalizer {
    my $class = shift;
    my $value = param_defined(shift);

    $value =~ s#\bany\b#any#ig;
    $value =~ s#\bLAN\b#LAN#ig;
    return $class->SUPER::text_normalizer($value);
}

sub text_trigger {
    my $self = shift;

    my $text = lc($self->text());

    if ($text =~ m#^(?:any|\*)$#) {
        $self->_repr([ $self->_repr_map("0.0.0.0/0") ]);
    } elsif ($text eq "lan") {
        # WARNING: this acts like a cache
        my %networks = AdvPowerMgr::System->get_networks();
        my @res = ();
        while (my ($net, $mask) = each %networks) {
            # Already in hex form
            push(@res, [ $net, $mask ]);
        }
        $self->_repr(\@res);
    } else {
        $self->SUPER::text_trigger();
    }
}

1;

# vi:set autoindent tabstop=8 shiftwidth=4 softtabstop=4 expandtab:
