# Copyright (c) 2017 Renaud METRICH
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products
#    derived from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# IPADDR_SPEC := [ RHOST[S] ] IPADDRS
# PORT_SPEC := [ LPORT[S] ] PORTS
# TUPLE := ( IPADDR_SPEC [ PORT_SPEC ] | [ IPADDR_SPEC ] PORT_SPEC )
package AdvPowerMgr::Plugin::PostShutdown::Connections::Tuple;

use strict;
use warnings;

use base qw( AdvPowerMgr::Plugin::GenericText );

use AdvPowerMgr::Errors;
use AdvPowerMgr::Plugin::PostShutdown::Connections::CIDR;
use AdvPowerMgr::Plugin::PostShutdown::Connections::Port;
use AdvPowerMgr::Utils qw( orregex );

sub _text_validator_i18n_msg {
    return 'connections.unexpected_tuple_string';
}

our $ipaddr_spec_re = "(?:RHOSTS?\\s+)?" . AdvPowerMgr::Plugin::PostShutdown::Connections::CIDR->_regex();
our $ipaddr_spec_matcher = "(?:RHOSTS?\\s+)?(" . AdvPowerMgr::Plugin::PostShutdown::Connections::CIDR->_regex() . ")";
our $port_spec_re = "(?:LPORTS?\\s+)?" . AdvPowerMgr::Plugin::PostShutdown::Connections::Port->_regex();
our $port_spec_matcher = "(?:LPORTS?\\s+)?(" . AdvPowerMgr::Plugin::PostShutdown::Connections::Port->_regex() . ")";
# Use a or to catch non-empty only
our $regex = orregex("$ipaddr_spec_re(?:\\s+$port_spec_re)?", "(?:$ipaddr_spec_re\\s+)?$port_spec_re");
our $regex_matcher = orregex("$ipaddr_spec_matcher(?:\\s+$port_spec_matcher)?", "(?:$ipaddr_spec_matcher\\s+)?$port_spec_matcher");

sub _regex {
    my $class = shift;

    return $regex;
}

sub _repr_map {
    my $class = shift;
    my $item = param_defined(shift);

    # uncoverable branch true
    $item =~ qr/^$regex_matcher$/i or die_ie(sprintf("%s: no match!", __PACKAGE__));
    my ($v1, $v2) = (defined($1) ? $1 : $3, defined($2) ? $2 : $4);

    return {
        cidr => AdvPowerMgr::Plugin::PostShutdown::Connections::CIDR->new(defined($v1) ? $v1 : "*"),
        port => AdvPowerMgr::Plugin::PostShutdown::Connections::Port->new(defined($v2) ? $v2 : "*")
    };
}

sub is_a_match {
    my $self = shift;
    my $r_ip = param_defined(shift);
    my $lport = param_defined(shift);

    foreach my $item (@{$self->_repr()}) {
        return 1 if $item->{cidr}->is_a_match($r_ip) && $item->{port}->is_a_match($lport);
    }

    return 0;
}

sub text_normalizer {
    my $class = shift;
    my $value = param_defined(shift);

    my @res = ();

    # uncoverable branch true
    $value =~ qr/^$regex_matcher$/i or die_ie(sprintf("%s: no match!", __PACKAGE__));
    my ($v1, $v2) = (defined($1) ? $1 : $3, defined($2) ? $2 : $4);
    if (defined($v1)) {
        push(@res, "rhosts") if $value =~ m#\bRHOSTS?\b#i;
        push(@res, AdvPowerMgr::Plugin::PostShutdown::Connections::CIDR->text_normalizer($v1));
    }
    if (defined($v2)) {
        push(@res, "lports") if $value =~ m#\bLPORTS?\b#i;
        push(@res, AdvPowerMgr::Plugin::PostShutdown::Connections::Port->text_normalizer($v2));
    }
    return join(" ", @res);
}

# TUPLE_LIST := LIST(TUPLE)
package AdvPowerMgr::Plugin::PostShutdown::Connections::TupleList;

use strict;
use warnings;

use base qw( AdvPowerMgr::Plugin::PostShutdown::Connections::Tuple );

use AdvPowerMgr::Errors;
use AdvPowerMgr::Utils qw( list list_splitter uniq );

sub _regex {
    my $class = shift;
    return list($class->SUPER::_regex());
}

sub text_normalizer {
    my $class = shift;
    my $value = param_defined(shift);

    my @res = ();
    my @list = list_splitter($class->SUPER::_regex(), $value);
    foreach my $item (@list) {
        push(@res, AdvPowerMgr::Plugin::PostShutdown::Connections::Tuple->text_normalizer($item));
    }
    return join(", ", @res);
}

sub text_trigger {
    my $self = shift;

    $self->_repr([ $self->list_split_and_map($self->SUPER::_regex()) ]);
}

# LINE := TUPLE_LIST [ EXCEPT TUPLE_LIST ]
package AdvPowerMgr::Plugin::PostShutdown::Connections::Line;

use strict;
use warnings;

use base qw( AdvPowerMgr::Plugin::GenericText );

use AdvPowerMgr::Errors;

sub _text_validator_i18n_msg {
    return 'connections.unexpected_line_string';
}

our $tlist = AdvPowerMgr::Plugin::PostShutdown::Connections::TupleList->_regex();
our $regex =  "$tlist(?:\\s+EXCEPT\\s+$tlist)?";
our $regex_matcher = qr/^($tlist)(?:\s+EXCEPT\s+($tlist))?$/i;

sub _regex {
    my $class = shift;

    return $regex;
}

sub _repr_map {
    my $class = shift;
    my $item = param_defined(shift);

    my @res = ();

    # uncoverable branch true
    $item =~ $regex_matcher or die_ie(sprintf("%s: no match!", __PACKAGE__));
    my ($v1, $v2) = ($1, $2);

    push(@res, AdvPowerMgr::Plugin::PostShutdown::Connections::TupleList->new($v1));
    push(@res, AdvPowerMgr::Plugin::PostShutdown::Connections::TupleList->new($v2)) if defined($v2);

    return @res;
}

sub is_a_match {
    my $self = shift;
    my $r_ip = param_defined(shift);
    my $l_port = param_defined(shift);

    if (@{$self->_repr()} == 2) {
        return 0 if $self->_repr()->[1]->is_a_match($r_ip, $l_port);
    }
    return $self->_repr()->[0]->is_a_match($r_ip, $l_port);
}

sub text_normalizer {
    my $class = shift;
    my $value = param_defined(shift);

    my @res = ();

    # uncoverable branch true
    $value =~ $regex_matcher or die_ie(sprintf("%s: no match!", __PACKAGE__));
    my ($v1, $v2) = ($1, $2);
    push(@res, AdvPowerMgr::Plugin::PostShutdown::Connections::TupleList->text_normalizer($v1));
    if (defined($v2)) {
        push(@res, "EXCEPT");
        push(@res, AdvPowerMgr::Plugin::PostShutdown::Connections::TupleList->text_normalizer($v2));
    }
    return  join("  ", @res);
}

1;

# vi:set autoindent tabstop=8 shiftwidth=4 softtabstop=4 expandtab:
