# Copyright (c) 2017 Renaud METRICH
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products
#    derived from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# SINGLE_PORT := ( 1 .. 65535 )
package AdvPowerMgr::Plugin::PostShutdown::Connections::SinglePort;

use strict;
use warnings;

use base qw( AdvPowerMgr::Plugin::GenericText );

use AdvPowerMgr::Errors;

    # 1 -> 65535 with or without leading zeros
our $regex = "(?:0?0?0?0?[1-9]|0?0?0?[1-9][0-9]|0?0?[1-9][0-9]{2}|0?[1-9][0-9]{3}|[1-5][0-9]{4}|6[0-4][0-9]{3}|65[0-4][0-9]{2}|655[0-2][0-9]|6553[0-5])";

sub _text_validator_i18n_msg {
    return 'connections.unexpected_port_string';
}

sub _regex {
    my $class = shift;

    return $regex;
}

sub _repr_map {
    my $class = shift;
    my $item = param_defined(shift);

    return int($item);
}

sub is_a_match {
    my $self = shift;
    my $value = param_defined(shift);

    unless ($value =~ m#^$regex$#) {
        my (undef, $filename, $line) = caller(1);
        die_ie("unexpected value '$value'", $filename, $line);
    }
    $value = int($value);
    return scalar(grep { /^$value$/ } @{$self->_repr()}) ? 1 : 0;
}

sub text_normalizer {
    my $class = shift;
    my $value = param_defined(shift);

    my $regex = __PACKAGE__->_regex();
    $value =~ s#\b($regex)\b#int($1)#ieg;

    return $class->SUPER::text_normalizer($value);
}

# PORT_LRANGE := LRANGE(SINGLE_PORT)
package AdvPowerMgr::Plugin::PostShutdown::Connections::PortLRange;

use strict;
use warnings;

use base qw( AdvPowerMgr::Plugin::PostShutdown::Connections::SinglePort );

use AdvPowerMgr::Utils qw( lrange uniqnumsort );

sub _regex {
    my $class = shift;
    return lrange($class->SUPER::_regex());
}

sub text_trigger {
    my $self = shift;

    $self->_repr([ uniqnumsort($self->lrange_split_and_map($self->SUPER::_regex())) ]);
}

sub _repr_map_range {
    my ($class, $first, $last) = @_;

    my @res = ();

    ($first, $last) = ($class->_repr_map($first), $class->_repr_map($last));
    if ($first <= $last) {
        push(@res, ($first .. $last));
    } else {
        push(@res, ($first .. 65535));
        push(@res, (1 .. $last));
    }
    
    return @res;
}

# PORT := ( * | any | PORT_LRANGE )
package AdvPowerMgr::Plugin::PostShutdown::Connections::Port;

use strict;
use warnings;

use base qw( AdvPowerMgr::Plugin::PostShutdown::Connections::PortLRange );

use AdvPowerMgr::Errors;
use AdvPowerMgr::Utils qw( orregex );

sub _regex {
    my $class = shift;
    return orregex("\\*", "any", $class->SUPER::_regex());
}

sub text_normalizer {
    my $class = shift;
    my $value = param_defined(shift);

    $value =~ s#\bany\b#any#ig;
    return $class->SUPER::text_normalizer($value);
}

sub text_trigger {
    my $self = shift;

    if (lc($self->text()) =~ m#^(?:any|\*)$#) {
        $self->_repr([ $self->_repr_map_range(1, 65535) ]);
    } else {
        $self->SUPER::text_trigger();
    }
}

1;

# vi:set autoindent tabstop=8 shiftwidth=4 softtabstop=4 expandtab:
