# Copyright (c) 2010-2017 Renaud METRICH
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products
#    derived from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package AdvPowerMgr::Plugin::Program;

use strict;
use warnings;

use base qw( AdvPowerMgr::Plugin );

use JSON qw();
use Time::Local qw( timelocal );

use AdvPowerMgr::Date;
use AdvPowerMgr::Errors;
use AdvPowerMgr::I18N qw( i18n );
use AdvPowerMgr::Log;
use AdvPowerMgr::Utils qw( min_ts max_ts remove_aref uniq );
use AdvPowerMgr::Plugin::Program::Date;
use AdvPowerMgr::Plugin::Program::Time;

__PACKAGE__->_mk_accessors(qw( list _repr ));

# TODO Create subclass for line parsing

#
# COMMENTS := ( # | @ | - ) TEXT
#
# LINE := TEXT1 AT TEXT2 EXCEPT TEXT3
#
# RE := DATE_SPEC AT TIME_SPEC [ EXCEPT DATE_SPEC ]
#
our $comments_re = __PACKAGE__->comments_re();
our $line_re = qr/^\s*(.*?)\s+AT\s+(.*?)(?:\s+EXCEPT\s+(.*?)\s*)?$/i;
our $regex_matcher = "^\\s*(" . AdvPowerMgr::Plugin::Program::Date->_regex() . ")\\s+AT\\s+(" . AdvPowerMgr::Plugin::Program::Time->_regex() . ")(?:\\s+EXCEPT\\s+(" . AdvPowerMgr::Plugin::Program::Date->_regex() . "))?\\s*\$";
our $re = qr/$regex_matcher/i;

sub list_validator {
    my $self = shift;
    my $value = shift;

    foreach my $item (@$value) {
        next if $item =~ $comments_re;
        
        unless ($item =~ $re) {
            # Try to find out which component is wrong
            die i18n('program.invalid_line', $item) unless $item =~ $line_re;
            AdvPowerMgr::Plugin::Program::Date->new($1);
            AdvPowerMgr::Plugin::Program::Time->new($2);
            # uncoverable branch false
            AdvPowerMgr::Plugin::Program::Date->new($3) if defined($3);
            # uncoverable statement
            die i18n('program.invalid_line', $item);
        }
    }

    return $value;
}

sub list_normalizer {
    my $self = shift;
    my $value = shift;

    my @res = ();

    foreach my $item (@$value) {
        if ($item =~ $comments_re) {
            $item =~ s#^\s+$##;
            push(@res, $item);
            next;
        }
        
        # uncoverable branch true
        $item =~ $re or die_ie(sprintf("%s: no match!", __PACKAGE__));
        my $date = AdvPowerMgr::Plugin::Program::Date->new($1, { disable_normalizers => 0 })->text();
        my $time = AdvPowerMgr::Plugin::Program::Time->new($2, { disable_normalizers => 0 })->text();
        if (defined($3)) {
            my $except = AdvPowerMgr::Plugin::Program::Date->new($3, { disable_normalizers => 0 })->text();
            push(@res, "$date  AT  $time  EXCEPT  $except");
        } else {
            push(@res, "$date  AT  $time");
        }
    }

    return \@res;
}

sub list_trigger {
    my $self = shift;

    my @res = ();

    foreach my $item (@{$self->list()}) {
        next if $item =~ $comments_re; # skip comments

        # uncoverable branch true
        $item =~ $re or die_ie(sprintf("%s: no match!", __PACKAGE__));
        my %r = (
            date => AdvPowerMgr::Plugin::Program::Date->new($1),
            time => AdvPowerMgr::Plugin::Program::Time->new($2),
            except => defined($3) ? AdvPowerMgr::Plugin::Program::Date->new($3) : undef
        );
        push(@res, { %r });
    }

    $self->_repr(\@res);
}

sub new {
    my $class = shift;
    my $params = defined($_[0]) ? { %{$_[0]} } : {}; shift;

    debug_func_enter($params);

    if (defined($params->{list}) && defined($params->{program})) {
        my (undef, $filename, $line) = caller(1);
        die_ie("'program' and 'list' are mutually exclusive", $filename, $line);
    }

    $params->{version} = 2;
    $params->{disable_normalizers} = 0 unless defined($params->{disable_normalizers}); 
    $params->{list} = [] unless defined($params->{list});
    $params->{list} = [ split(/\n/, $params->{program}) ] if defined($params->{program});

    if (defined($params->{objtype}) && $params->{objtype} eq "AdvPowerMgr::Plugin::Program") {
        # Upgrade from 3.x, x <= 2
        delete $params->{objtype};
    }

    my $self = $class->SUPER::new($params);
    bless($self, $class);
    $self->list($params->{list});

    return $self;
}

sub to_hash {
    my $self = shift;
    my $options = shift || {};

    my $ret = $self->SUPER::to_hash($options);
    $ret->{program} = join("\n", @{$self->list()});
    if ($options->{for_json}) {
        $ret->{disable_normalizers} = $self->disable_normalizers() ? JSON->true() : JSON->false();
    } else {
        $ret->{disable_normalizers} = $self->disable_normalizers();
    }

    return $ret;
}

sub compute_ts {
    my $self = shift;
    my $when = shift || time();

    return (wantarray ? undef : 0) if !$self->enabled();

    my $res = undef;
    my $startdate = AdvPowerMgr::Date->new($when);
    my $daysahead = 7;

    while (1) {
        # Initialize to 'undef' so that objects such as SingleDateWithYear can
        # eventually return 'undef', making the loop stop
        $res = undef;

        die_ie("Date too far in the future!") if $daysahead >= 3600;

        foreach my $item (@{$self->_repr()}) {
            my $dates = $item->{date}->compute_ts($startdate, $daysahead);
            my @times = @{$item->{time}->_repr()};

            next unless defined($dates);
            # We have a date, or at least one in the future ($dates = []), hence initialize $res
            $res = [] unless defined($res);

            if (defined($item->{except})) {
                my $except = $item->{except}->compute_ts($startdate, $daysahead);
                $dates = remove_aref($dates, $except);
            }
            foreach my $d (@{$dates}) {
                my (undef, undef, undef, $dm, $mo, $yo) = localtime($d);
                foreach my $t (@times) {
                    $res = [] unless defined($res);

                    push(@{$res}, timelocal(0, $t->{m}, $t->{h}, $dm, $mo, $yo));
                }
            }
        }
        # No need to iterate if no value will ever be returned
        last unless defined($res);
        $res = [ uniq(sort { $a <=> $b } @$res) ];
        while (@$res && $res->[0] <= $when) {
            shift(@$res);
        }
        # No need to iterate if we found some value
        last if @$res;
        $startdate += $daysahead;
        $daysahead *= 2;
    }

    return (defined($res) ? @$res : $res) if wantarray;

    return 0 if !defined($res) || !@$res;
    return $res->[0];
}

1;

# vi:set autoindent tabstop=8 shiftwidth=4 softtabstop=4 expandtab:
