# Copyright (c) 2017 Renaud METRICH
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products
#    derived from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package AdvPowerMgr::Plugin::Program::Date::Generic;

use strict;
use warnings;

use base qw( AdvPowerMgr::Plugin::GenericText );

use AdvPowerMgr::I18N qw( i18n );

sub _text_validator_i18n_msg {
    return 'program.unexpected_date_string';
}

# MONTHDAY := ( 1 .. 31 )
package AdvPowerMgr::Plugin::Program::Date::Monthday;

use strict;
use warnings;

use base qw( AdvPowerMgr::Plugin::Program::Date::Generic );

use AdvPowerMgr::Date;
use AdvPowerMgr::Errors;

sub _regex {
    my $class = shift;
    return "(?:0?[1-9]|[1-2][0-9]|3[0-1])";
}

sub _repr_map {
    my $class = shift;
    my $item = param_defined(shift);

    return int($item);
}

sub compute_ts {
    my $self = shift;
    my $when = AdvPowerMgr::Date->new(defined($_[0]) ? $_[0] : time()); shift;
    my $daysahead = shift; defined($daysahead) or $daysahead = 7;

    my @res = ();

    for (my $i = 0; $i <= $daysahead; $i++) {
        my $day = $when->day();
        push(@res, $when->ts()) if grep { /^$day$/ } @{$self->_repr()};
        $when++;
    }
        
    return \@res;
}

sub text_normalizer {
    my $class = shift;
    my $value = param_defined(shift);

    my $regex = __PACKAGE__->_regex();
    $value =~ s#\b($regex)\b#int($1)#ieg;

    return $class->SUPER::text_normalizer($value);
}

# MONTHDAY_LRANGE := LRANGE(MONTHDAY)
package AdvPowerMgr::Plugin::Program::Date::MonthdayLRange;

use strict;
use warnings;

use base qw( AdvPowerMgr::Plugin::Program::Date::Monthday );

use AdvPowerMgr::Utils qw( lrange uniqnumsort );

sub _regex {
    my $class = shift;
    return lrange($class->SUPER::_regex());
}

sub _repr_map_range {
    my ($class, $first, $last) = @_;

    my @res = ();

    ($first, $last) = ($class->_repr_map($first), $class->_repr_map($last));
    if ($first <= $last) {
        push(@res, ($first .. $last));
    } else {
        push(@res, ($first .. 31));
        push(@res, (1 .. $last));
    }
    
    return @res;
}

sub text_trigger {
    my $self = shift;

    $self->_repr([ uniqnumsort($self->lrange_split_and_map($self->SUPER::_regex())) ]);
}

# MONTH := ( jan | feb | mar | apr | may | jun | jul | aug | sep | oct | nov | dec )
package AdvPowerMgr::Plugin::Program::Date::Month;

use strict;
use warnings;

use base qw( AdvPowerMgr::Plugin::Program::Date::Generic );

use AdvPowerMgr::Date;
use AdvPowerMgr::Errors;
use AdvPowerMgr::Utils qw( orregex );

our @month_arr = ( undef, "jan", "feb", "mar", "apr", "may", "jun", "jul", "aug", "sep", "oct", "nov", "dec" );
our %month_map = (); for (my $i = 1; $i <= 12; $i++) { $month_map{$month_arr[$i]} = $i; };

sub _repr_map {
    my $class = shift;
    my $item = param_defined(shift);

    return $month_map{lc($item)};
}

sub _regex {
    my $class = shift;
    return orregex(@month_arr[1..12]);
}

sub compute_ts {
    my $self = shift;
    my $when = AdvPowerMgr::Date->new(defined($_[0]) ? $_[0] : time()); shift;
    my $daysahead = shift; defined($daysahead) or $daysahead = 7;

    my @res = ();

    for (my $i = 0; $i <= $daysahead; $i++) {
        my $month = $when->month();
        push(@res, $when->ts()) if grep { /^$month$/ } @{$self->_repr()};
        $when++;
    }
        
    return \@res;
}

sub text_normalizer {
    my $class = shift;
    my $value = param_defined(shift);

    my $regex = __PACKAGE__->_regex();
    $value =~ s#\b($regex)\b#ucfirst(lc($1))#ieg;

    return $class->SUPER::text_normalizer($value);
}

# MONTH_LRANGE := LRANGE(MONTH)
package AdvPowerMgr::Plugin::Program::Date::MonthLRange;

use strict;
use warnings;

use base qw( AdvPowerMgr::Plugin::Program::Date::Month );

use AdvPowerMgr::Utils qw( lrange uniqnumsort );

sub _regex {
    my $class = shift;
    return lrange($class->SUPER::_regex());
}

sub _repr_map_range {
    my ($class, $first, $last) = @_;

    my @res = ();

    ($first, $last) = ($class->_repr_map($first), $class->_repr_map($last));
    if ($first <= $last) {
        push(@res, ($first .. $last));
    } else {
        push(@res, ($first .. 12));
        push(@res, (1 .. $last));
    }
    
    return @res;
}

sub text_trigger {
    my $self = shift;

    $self->_repr([ uniqnumsort($self->lrange_split_and_map($self->SUPER::_regex())) ]);
}

# NUMERAL := ( first | second | third | fourth | fifth | last )
package AdvPowerMgr::Plugin::Program::Date::Numeral;

use strict;
use warnings;

use base qw( AdvPowerMgr::Plugin::Program::Date::Generic );

use AdvPowerMgr::Date;
use AdvPowerMgr::Errors;
use AdvPowerMgr::Utils qw( orregex );

our @numeral_arr = ( "last", "first", "second", "third", "fourth", "fifth" );
our %numeral_map = (); for (my $i = 0; $i <= 5; $i++) { $numeral_map{$numeral_arr[$i]} = $i; };

sub _repr_map {
    my $self = shift;
    my $item = param_defined(shift);

    return $numeral_map{lc($item)};
}

sub _regex {
    my $class = shift;
    return orregex(@numeral_arr[0..5]);
}

sub compute_ts {
    my $self = shift;
    my $when = AdvPowerMgr::Date->new(defined($_[0]) ? $_[0] : time()); shift;
    my $daysahead = shift; defined($daysahead) or $daysahead = 7;

    my @res = ();

    # For 'last'
    my $haslast = scalar(grep { /^0$/ } @{$self->_repr()}) ? 1 : 0;
    my ($month, $laststart) = (0, undef);

    for (my $i = 0; $i <= $daysahead; $i++) {
        my $week = int(($when->day() - 1) / 7) + 1;
        push(@res, $when->ts()) if grep { /^$week$/ } @{$self->_repr()};
        if ($haslast) {
            if ($when->month() != $month) {
                if ($when->month() == 12) {
                    $laststart = (AdvPowerMgr::Date->new({
                        year => $when->year() + 1,
                        month => 1,
                        day => 1
                    }) - 7)->day();
                } else {
                    $laststart = (AdvPowerMgr::Date->new({
                        year => $when->year(),
                        month => $when->month() + 1,
                        day => 1
                    }) - 7)->day();
                }
                $month = $when->month();
            }
            push(@res, $when->ts()) if $when->day() >= $laststart;
        }
        $when++;
    }
        
    return \@res;
}

sub text_normalizer {
    my $class = shift;
    my $value = param_defined(shift);

    my $regex = __PACKAGE__->_regex();
    $value =~ s#\b($regex)\b#lc($1)#ieg;

    return $class->SUPER::text_normalizer($value);
}

# NUMERAL_LRANGE := LRANGE(NUMERAL)
package AdvPowerMgr::Plugin::Program::Date::NumeralLRange;

use strict;
use warnings;

use base qw( AdvPowerMgr::Plugin::Program::Date::Numeral );

use AdvPowerMgr::I18N qw( i18n );
use AdvPowerMgr::Utils qw( lrange uniqnumsort );

sub _regex {
    my $class = shift;
    return lrange($class->SUPER::_regex());
}

sub _repr_map_range {
    my ($class, $first_text, $last_text) = @_;

    my @res = ();

    my ($first, $last) = ($class->_repr_map($first_text), $class->_repr_map($last_text));
    if ($first != 0 && $last != 0) {
        if ($first <= $last) {
            push(@res, ($first .. $last));
        } else {
            push(@res, ($first .. 5));
            push(@res, (1 .. $last));
        }
    } elsif ($first == 0) {
        if ($last != 0) {
            # last - something: there is ambiguity when something = fourth or fifth
            die i18n('program.invalid_numeral_range', "$first_text-$last_text");
        }
        push(@res, 0);
    } else {
        # something - last == something .. fifth
        push(@res, ($first .. 5));
    }

    return @res;
}

sub text_trigger {
    my $self = shift;

    $self->_repr([ uniqnumsort($self->lrange_split_and_map($self->SUPER::_regex())) ]);
}

# WEEKDAY := ( mon | tue | wed | thr | fri | sat | sun )
package AdvPowerMgr::Plugin::Program::Date::Weekday;

use strict;
use warnings;

use base qw( AdvPowerMgr::Plugin::Program::Date::Generic );

use AdvPowerMgr::Date;
use AdvPowerMgr::Errors;
use AdvPowerMgr::Utils qw( orregex );

our @weekday_arr = ( undef, "mon", "tue", "wed", "thr", "fri", "sat", "sun" );
our %weekday_map = (); for (my $i = 1; $i <= 7; $i++) { $weekday_map{$weekday_arr[$i]} = $i; };

sub _repr_map {
    my $self = shift;
    my $item = param_defined(shift);

    return $weekday_map{lc($item)};
}

sub _regex {
    my $class = shift;
    return orregex(@weekday_arr[1..7]);
}

sub compute_ts {
    my $self = shift;
    my $when = AdvPowerMgr::Date->new(defined($_[0]) ? $_[0] : time()); shift;
    my $daysahead = shift; defined($daysahead) or $daysahead = 7;

    my @res = ();

    for (my $i = 0; $i <= $daysahead; $i++) {
        my $weekday = $when->weekday();

        push(@res, $when->ts()) if grep { /^$weekday$/ } @{$self->_repr()};
        $when++;
    }
        
    return \@res;
}

sub text_normalizer {
    my $class = shift;
    my $value = param_defined(shift);

    my $regex = __PACKAGE__->_regex();
    $value =~ s#\b($regex)\b#ucfirst(lc($1))#ieg;

    return $class->SUPER::text_normalizer($value);
}

# WEEKDAY_LRANGE := LRANGE(WEEKDAY)
package AdvPowerMgr::Plugin::Program::Date::WeekdayLRange;

use strict;
use warnings;

use base qw( AdvPowerMgr::Plugin::Program::Date::Weekday );

use AdvPowerMgr::Errors;
use AdvPowerMgr::Utils qw( lrange uniqnumsort );

sub _regex {
    my $class = shift;
    return lrange($class->SUPER::_regex());
}

sub _repr_map_range {
    my ($class, $first, $last) = @_;

    my @res = ();

    ($first, $last) = ($class->_repr_map($first), $class->_repr_map($last));
    if ($first <= $last) {
        push(@res, ($first .. $last));
    } else {
        push(@res, ($first .. 7));
        push(@res, (1 .. $last));
    }
    
    return @res;
}

sub text_trigger {
    my $self = shift;

    $self->_repr([ uniqnumsort($self->lrange_split_and_map($self->SUPER::_regex())) ]);
}

# WEEKNUM := ( 1 .. 53 )
package AdvPowerMgr::Plugin::Program::Date::Weeknum;

use strict;
use warnings;

use base qw( AdvPowerMgr::Plugin::Program::Date::Generic );

use AdvPowerMgr::Date;
use AdvPowerMgr::Errors;

sub _regex {
    my $class = shift;
    return "(?:0?[1-9]|[1-4][0-9]|5[0-3])";
}

sub _repr_map {
    my $class = shift;
    my $item = param_defined(shift);

    return int($item);
}

sub compute_ts {
    my $self = shift;
    my $when = AdvPowerMgr::Date->new(defined($_[0]) ? $_[0] : time()); shift;
    my $daysahead = shift; defined($daysahead) or $daysahead = 7;

    my @res = ();

    for (my $i = 0; $i <= $daysahead; $i++) {
        my $weeknum = $when->weeknum();

        push(@res, $when->ts()) if grep { /^$weeknum$/ } @{$self->_repr()};
        $when++;
    }
        
    return \@res;
}

sub text_normalizer {
    my $class = shift;
    my $value = param_defined(shift);

    my $regex = __PACKAGE__->_regex();
    $value =~ s#\b($regex)\b#int($1)#eg;

    return $class->SUPER::text_normalizer($value);
}

# WEEKNUM_LRANGE := LRANGE(WEEKNUM)
package AdvPowerMgr::Plugin::Program::Date::WeeknumLRange;

use strict;
use warnings;

use base qw( AdvPowerMgr::Plugin::Program::Date::Weeknum );

use AdvPowerMgr::Utils qw( lrange uniqnumsort );

sub _regex {
    my $class = shift;
    return lrange($class->SUPER::_regex());
}

sub _repr_map_range {
    my ($class, $first, $last) = @_;

    my @res = ();

    ($first, $last) = ($class->_repr_map($first), $class->_repr_map($last));
    if ($first <= $last) {
        push(@res, ($first .. $last));
    } else {
        push(@res, ($first .. 53));
        push(@res, (1 .. $last));
    }
    
    return @res;
}

sub text_trigger {
    my $self = shift;

    $self->_repr([ uniqnumsort($self->lrange_split_and_map($self->SUPER::_regex())) ]);
}

# YEARDAY := ( 1 .. 366 )
package AdvPowerMgr::Plugin::Program::Date::Yearday;

use strict;
use warnings;

use base qw( AdvPowerMgr::Plugin::Program::Date::Generic );

use AdvPowerMgr::Date;
use AdvPowerMgr::Errors;
use AdvPowerMgr::Utils qw( orregex );

sub _regex {
    my $class = shift;
    return "(?:0?0?[1-9]|0?[1-9][0-9]|[1-2][0-9][0-9]|3[0-5][0-9]|36[0-6])";
}

sub _repr_map {
    my $class = shift;
    my $item = param_defined(shift);

    return int($item);
}

sub compute_ts {
    my $self = shift;
    my $when = AdvPowerMgr::Date->new(defined($_[0]) ? $_[0] : time()); shift;
    my $daysahead = shift; defined($daysahead) or $daysahead = 7;

    my @res = ();

    for (my $i = 0; $i <= $daysahead; $i++) {
        my $yearday = $when->yearday();

        push(@res, $when->ts()) if grep { /^$yearday$/ } @{$self->_repr()};
        $when++;
    }
        
    return \@res;
}

sub text_normalizer {
    my $class = shift;
    my $value = param_defined(shift);

    my $regex = __PACKAGE__->_regex();
    $value =~ s#\b($regex)\b#int($1)#eg;

    return $class->SUPER::text_normalizer($value);
}

# YEARDAY_LRANGE := LRANGE(YEARDAY)
package AdvPowerMgr::Plugin::Program::Date::YeardayLRange;

use strict;
use warnings;

use base qw( AdvPowerMgr::Plugin::Program::Date::Yearday );

use AdvPowerMgr::Utils qw( lrange uniqnumsort );

sub _regex {
    my $class = shift;
    return lrange($class->SUPER::_regex());
}

sub _repr_map_range {
    my ($class, $first, $last) = @_;

    my @res = ();

    ($first, $last) = ($class->_repr_map($first), $class->_repr_map($last));
    if ($first <= $last) {
        push(@res, ($first .. $last));
    } else {
        push(@res, ($first .. 366));
        push(@res, (1 .. $last));
    }
    
    return @res;
}

sub text_trigger {
    my $self = shift;

    $self->_repr([ uniqnumsort($self->lrange_split_and_map($self->SUPER::_regex())) ]);
}

# YEAR := [2-9][0-9][0-9][0-9]
package AdvPowerMgr::Plugin::Program::Date::Year;

use strict;
use warnings;

use base qw( AdvPowerMgr::Plugin::Program::Date::Generic );

use AdvPowerMgr::Date;

sub _regex {
    return "[2-9][0-9][0-9][0-9]";
}

sub compute_ts {
    my $self = shift;
    my $when = AdvPowerMgr::Date->new(defined($_[0]) ? $_[0] : time()); shift;
    my $daysahead = shift; defined($daysahead) or $daysahead = 7;

    return undef if $when->year() > [ sort { $a <=> $b } @{$self->_repr()} ]->[-1];

    my @res = ();

    for (my $i = 0; $i <= $daysahead; $i++) {
        my $year = $when->year();

        push(@res, $when->ts()) if grep { /^$year$/ } @{$self->_repr()};
        $when++;
    }
        
    return \@res;
}

# YEAR_LRANGE := LRANGE(YEAR)
package AdvPowerMgr::Plugin::Program::Date::YearLRange;

use strict;
use warnings;

use base qw( AdvPowerMgr::Plugin::Program::Date::Year );

use AdvPowerMgr::I18N qw( i18n );
use AdvPowerMgr::Utils qw( lrange uniqnumsort );

sub _regex {
    my $class = shift;
    return lrange($class->SUPER::_regex());
}

sub _repr_map_range {
    my ($class, $first, $last) = @_;

    my @res = ();

    ($first, $last) = ($class->_repr_map($first), $class->_repr_map($last));
    if ($first <= $last) {
        push(@res, ($first .. $last));
    } else {
        die i18n('program.invalid_range', "$first-$last");
    }
    
    return @res;
}

sub text_trigger {
    my $self = shift;

    $self->_repr([ uniqnumsort($self->lrange_split_and_map($self->SUPER::_regex())) ]);
}

# MONTH_SELECTOR := ( monthly | MONTH_LRANGE )
package AdvPowerMgr::Plugin::Program::Date::MonthSelector;

use strict;
use warnings;

use base qw( AdvPowerMgr::Plugin::Program::Date::MonthLRange );

use AdvPowerMgr::Errors;
use AdvPowerMgr::Utils qw( orregex );

sub _regex {
    my $class = shift;
    return orregex("monthly", $class->SUPER::_regex());
}

sub text_normalizer {
    my $class = shift;
    my $value = param_defined(shift);

    $value =~ s#\bmonthly\b#monthly#ig;

    return $class->SUPER::text_normalizer($value);
}

sub text_trigger {
    my $self = shift;

    if (lc($self->text()) eq "monthly") {
        $self->_repr([ $self->_repr_map_range("jan", "dec") ]);
    } else {
        $self->SUPER::text_trigger();
    }
}

# NUMERALWEEKDAY_SELECTOR := LIST(NUMERAL_LRANGE WEEKDAY_LRANGE)
package AdvPowerMgr::Plugin::Program::Date::NumeralWeekdaySelector;

use strict;
use warnings;

use base qw( AdvPowerMgr::Plugin::Program::Date::Generic );

use AdvPowerMgr::Date;
use AdvPowerMgr::Errors;
use AdvPowerMgr::Utils qw( intersect_aref list list_splitter uniqnumsort );

our $regex = AdvPowerMgr::Plugin::Program::Date::NumeralLRange->_regex() . "\\s+" . AdvPowerMgr::Plugin::Program::Date::WeekdayLRange->_regex();
our $regex_matcher = "^(" . AdvPowerMgr::Plugin::Program::Date::NumeralLRange->_regex() . ")\\s+(" . AdvPowerMgr::Plugin::Program::Date::WeekdayLRange->_regex() . ")\$";

sub _regex {
    my $class = shift;
    return list($regex);
}

sub text_normalizer {
    my $class = shift;
    my $value = param_defined(shift);

    $value = AdvPowerMgr::Plugin::Program::Date::NumeralLRange->text_normalizer($value);
    $value = AdvPowerMgr::Plugin::Program::Date::WeekdayLRange->text_normalizer($value);

    return $class->SUPER::text_normalizer($value);
}

sub text_trigger {
    my $self = shift;

    my @list = list_splitter($regex, $self->text());
    my @res = ();

    foreach my $item (@list) {
        $item =~ m#$regex_matcher#i;
        push(@res, [ AdvPowerMgr::Plugin::Program::Date::NumeralLRange->new($1), AdvPowerMgr::Plugin::Program::Date::WeekdayLRange->new($2) ]);
    }

    $self->_repr(\@res);
}

sub compute_ts {
    my $self = shift;
    my $when = AdvPowerMgr::Date->new(defined($_[0]) ? $_[0] : time()); shift;
    my $daysahead = shift; defined($daysahead) or $daysahead = 7;

    my @res = ();

    foreach my $pair (@{$self->_repr()}) {
        my $res1 = $pair->[0]->compute_ts($when, $daysahead);
        my $res2 = $pair->[1]->compute_ts($when, $daysahead);
        my $res = intersect_aref($res1, $res2);
        push(@res, @{$res});
    }

    @res = uniqnumsort(@res);
    return \@res;
}

# WEEK_SELECTOR := ( weekly | week WEEKNUM_LRANGE )
package AdvPowerMgr::Plugin::Program::Date::WeekSelector;

use strict;
use warnings;

use base qw( AdvPowerMgr::Plugin::Program::Date::WeeknumLRange );

use AdvPowerMgr::Errors;
use AdvPowerMgr::Utils qw( orregex );

sub _regex {
    my $class = shift;
    return orregex("weekly", "week " . $class->SUPER::_regex());
}

sub text_normalizer {
    my $class = shift;
    my $value = param_defined(shift);

    $value =~ s#\bweekly\b#weekly#ig;
    $value =~ s#\bweek\b#week#ig;

    return $class->SUPER::text_normalizer($value);
}

sub text_trigger {
    my $self = shift;

    if (lc($self->text()) eq "weekly") {
        $self->_repr([ 1 .. 53 ]);
    } else {
        my $regex = AdvPowerMgr::Plugin::Program::Date::WeeknumLRange->_regex();
        $self->text() =~ m#^week\s+($regex)$#i;
        $self->_repr(AdvPowerMgr::Plugin::Program::Date::WeeknumLRange->new($1)->_repr());
    }
}

# WEEKDAY_SELECTOR := ( daily | WEEKDAY_LRANGE )
package AdvPowerMgr::Plugin::Program::Date::WeekdaySelector;

use strict;
use warnings;

use base qw( AdvPowerMgr::Plugin::Program::Date::WeekdayLRange );

use AdvPowerMgr::Errors;
use AdvPowerMgr::Utils qw( orregex );

sub _regex {
    my $class = shift;
    return orregex("daily", $class->SUPER::_regex());
}

sub text_normalizer {
    my $class = shift;
    my $value = param_defined(shift);

    $value =~ s#\bdaily\b#daily#ig;
    return $class->SUPER::text_normalizer($value);
}

sub text_trigger {
    my $self = shift;

    if (lc($self->text()) eq "daily") {
        $self->_repr([ $self->_repr_map_range("mon", "sun") ]);
    } else {
        $self->SUPER::text_trigger();
    }
}

# SINGLE_DATE_NOYEAR := MONTH MONTHDAY
package AdvPowerMgr::Plugin::Program::Date::SingleDateNoYear;

use strict;
use warnings;

use base qw( AdvPowerMgr::Plugin::Program::Date::Generic );

use AdvPowerMgr::Date;
use AdvPowerMgr::Errors;
use AdvPowerMgr::Utils qw( intersect_aref );

our $regex = AdvPowerMgr::Plugin::Program::Date::Month->_regex() . "\\s+" . AdvPowerMgr::Plugin::Program::Date::Monthday->_regex();
our $regex_matcher = "^(" . AdvPowerMgr::Plugin::Program::Date::Month->_regex() . ")\\s+(" . AdvPowerMgr::Plugin::Program::Date::Monthday->_regex() . ")\$";

sub _regex {
    my $class = shift;
    return $regex;
}

sub text_normalizer {
    my $class = shift;
    my $value = param_defined(shift);

    my @res = ();
    $value =~ m#$regex_matcher#i;
    push(@res, AdvPowerMgr::Plugin::Program::Date::Month->text_normalizer($1));
    push(@res, AdvPowerMgr::Plugin::Program::Date::Monthday->text_normalizer($2));

    return join(" ", @res);
}

sub text_trigger {
    my $self = shift;

    my @res = ();

    $self->text() =~ m#$regex_matcher#i;
    push(@res, AdvPowerMgr::Plugin::Program::Date::Month->new($1));
    push(@res, AdvPowerMgr::Plugin::Program::Date::Monthday->new($2));

    $self->_repr(\@res);
}

sub compute_ts {
    my $self = shift;
    my $when = AdvPowerMgr::Date->new(defined($_[0]) ? $_[0] : time()); shift;
    my $daysahead = shift; defined($daysahead) or $daysahead = 7;

    my ($month, $monthday) = @{$self->_repr()};
    my $res1 = $month->compute_ts($when, $daysahead);
    my $res2 = $monthday->compute_ts($when, $daysahead);
    my $res = intersect_aref($res1, $res2);

    return $res;
}

# SINGLE_DATE_WITHYEAR := MONTH MONTHDAY YEAR
package AdvPowerMgr::Plugin::Program::Date::SingleDateWithYear;

use strict;
use warnings;

use base qw( AdvPowerMgr::Plugin::Program::Date::Generic );

use AdvPowerMgr::Date;
use AdvPowerMgr::Errors;
use AdvPowerMgr::I18N qw( i18n );
use AdvPowerMgr::Utils qw( intersect_aref );

our $regex = AdvPowerMgr::Plugin::Program::Date::Month->_regex() . "\\s+" . AdvPowerMgr::Plugin::Program::Date::Monthday->_regex() . "\\s+" . AdvPowerMgr::Plugin::Program::Date::Year->_regex();
our $regex_matcher = "^(" . AdvPowerMgr::Plugin::Program::Date::Month->_regex() . ")\\s+(" . AdvPowerMgr::Plugin::Program::Date::Monthday->_regex() . ")\\s+(" . AdvPowerMgr::Plugin::Program::Date::Year->_regex() . ")\$";

sub _regex {
    my $class = shift;
    return $regex;
}

sub text_normalizer {
    my $class = shift;
    my $value = param_defined(shift);

    my @res = ();
    $value =~ m#$regex_matcher#i;
    push(@res, AdvPowerMgr::Plugin::Program::Date::Month->text_normalizer($1));
    push(@res, AdvPowerMgr::Plugin::Program::Date::Monthday->text_normalizer($2));
    push(@res, AdvPowerMgr::Plugin::Program::Date::Year->text_normalizer($3));

    return join(" ", @res);
}

sub text_trigger {
    my $self = shift;

    $self->text() =~ m#$regex_matcher#i;

    my ($month) = @{AdvPowerMgr::Plugin::Program::Date::Month->new($1)->_repr()};
    my ($day) = @{AdvPowerMgr::Plugin::Program::Date::Monthday->new($2)->_repr()};
    my ($year) = @{AdvPowerMgr::Plugin::Program::Date::Year->new($3)->_repr()};

    eval {
        $self->_repr(AdvPowerMgr::Date->new({ year => $year, month => $month, day => $day }));
    };
    die i18n('program.invalid_date', $self->text()) if $@;
}

sub compute_ts {
    my $self = shift;
    my $when = AdvPowerMgr::Date->new(defined($_[0]) ? $_[0] : time()); shift;
    my $daysahead = shift; defined($daysahead) or $daysahead = 7;

    return undef if $self->_repr() < $when;
    return [] if $self->_repr() > $when + $daysahead;
    return [ $self->_repr()->ts() ];
}

# DATE_SPEC := MONTH_SELECTOR MONTHDAY_LRANGE [ YEAR_LRANGE ]
package AdvPowerMgr::Plugin::Program::Date::DateSpec;

use strict;
use warnings;

use base qw( AdvPowerMgr::Plugin::Program::Date::Generic );

use AdvPowerMgr::Date;
use AdvPowerMgr::Errors;
use AdvPowerMgr::Utils qw( intersect_aref );

our $regex = AdvPowerMgr::Plugin::Program::Date::MonthSelector->_regex() . "\\s+" . AdvPowerMgr::Plugin::Program::Date::MonthdayLRange->_regex() . "(?:\\s+" . AdvPowerMgr::Plugin::Program::Date::YearLRange->_regex() . ")?";
our $regex_matcher = "^(" . AdvPowerMgr::Plugin::Program::Date::MonthSelector->_regex() . ")\\s+(" . AdvPowerMgr::Plugin::Program::Date::MonthdayLRange->_regex() . ")(?:\\s+(" . AdvPowerMgr::Plugin::Program::Date::YearLRange->_regex() . "))?\$";

sub _regex {
    my $class = shift;
    return $regex;
}

sub text_normalizer {
    my $class = shift;
    my $value = param_defined(shift);

    my @res = ();
    $value =~ m#$regex_matcher#i;
    push(@res, AdvPowerMgr::Plugin::Program::Date::MonthSelector->text_normalizer($1));
    push(@res, AdvPowerMgr::Plugin::Program::Date::MonthdayLRange->text_normalizer($2));
    push(@res, AdvPowerMgr::Plugin::Program::Date::YearLRange->text_normalizer($3)) if defined($3);

    return join(" ", @res);
}

sub text_trigger {
    my $self = shift;

    my @res = ();

    $self->text() =~ m#$regex_matcher#i;
    push(@res, AdvPowerMgr::Plugin::Program::Date::MonthSelector->new($1));
    push(@res, AdvPowerMgr::Plugin::Program::Date::MonthdayLRange->new($2));
    push(@res, AdvPowerMgr::Plugin::Program::Date::YearLRange->new($3)) if defined($3);

    $self->_repr(\@res);
}

sub compute_ts {
    my $self = shift;
    my $when = AdvPowerMgr::Date->new(defined($_[0]) ? $_[0] : time()); shift;
    my $daysahead = shift; defined($daysahead) or $daysahead = 7;

    my @arr = ();
    foreach my $obj (@{$self->_repr()}) {
        push(@arr, $obj->compute_ts($when, $daysahead));
    }

    return intersect_aref(@arr);
}

# DATERANGE_SPEC_NOYEAR = RANGE(SINGLE_DATE_NOYEAR)
package AdvPowerMgr::Plugin::Program::Date::DateRangeSpecNoYear;

use strict;
use warnings;

use base qw( AdvPowerMgr::Plugin::Program::Date::SingleDateNoYear );

use AdvPowerMgr::Date;
use AdvPowerMgr::Errors;
use AdvPowerMgr::Utils qw( intersect_aref range range_splitter uniqnumsort );

sub _regex {
    my $class = shift;
    return range(AdvPowerMgr::Plugin::Program::Date::SingleDateNoYear->_regex());
}

sub text_normalizer {
    my $class = shift;
    my $value = param_defined(shift);

    my ($first, $last) = range_splitter(AdvPowerMgr::Plugin::Program::Date::SingleDateNoYear->_regex(), $value);
    my $r1 = AdvPowerMgr::Plugin::Program::Date::SingleDateNoYear->text_normalizer($first);
    my $r2 = AdvPowerMgr::Plugin::Program::Date::SingleDateNoYear->text_normalizer($last);

    return $r1 eq $r2 ? $r1 : "$r1-$r2";
}

sub text_trigger {
    my $self = shift;

    my ($first, $last) = range_splitter(AdvPowerMgr::Plugin::Program::Date::SingleDateNoYear->_regex(), $self->text());
    my @res = ();

    push(@res, AdvPowerMgr::Plugin::Program::Date::SingleDateNoYear->new($first));
    push(@res, AdvPowerMgr::Plugin::Program::Date::SingleDateNoYear->new($last)) unless $last eq $first;

    $self->_repr(\@res);
}

sub compute_ts {
    my $self = shift;
    my $when = AdvPowerMgr::Date->new(defined($_[0]) ? $_[0] : time()); shift;
    my $daysahead = shift; defined($daysahead) or $daysahead = 7;

    my ($first, $last) = @{$self->_repr()};

    my $res1 = $first->compute_ts($when, $daysahead);
    return $res1 unless defined($last); # Single date, not a range

    # compute @dates = ( $when; $when + $daysahead )
    # compute @ylist (years list)
    #
    # if $first < $last
    #   foreach year (@ylist)
    #       @res += compute [ $first + year; $last + year ]
    # else
    #   foreach year (@ylist)
    #       @res += compute [ $first - 1year; $last ]
    #       @res += compute [ $first; $last + 1year ]
    #
    # intersect @dates, @res

    my @dates = ();
    my %years = ();
    for (my $i = 0; $i <= $daysahead; $i++) {
        my $date = AdvPowerMgr::Date->new($when + $i);
        $years{$date->year()} = 1;
        push(@dates, $date->ts());
    }
    my @ylist = keys(%years);

    my @res = ();
    my $first_ts = $first->compute_ts($when, 366)->[0];
    my $last_ts = $last->compute_ts($when, 366)->[0];
    $first = AdvPowerMgr::Date->new($first_ts);
    $last = AdvPowerMgr::Date->new($last_ts);
    $last->year($first->year());

    if ($first->ts() < $last->ts()) {
        foreach my $year (@ylist) {
            $first->year($year);
            $last->year($year);
            my $date = AdvPowerMgr::Date->new($first);
            while ($date <= $last) {
                push(@res, $date->ts());
                $date++;
            }
        }
    } else {
        foreach my $year (@ylist) {
            foreach my $offset (-1, 0) {
                $first->year($year + $offset);
                $last->year($year + $offset + 1);
                my $date = AdvPowerMgr::Date->new($first);
                while ($date <= $last) {
                    push(@res, $date->ts());
                    $date++;
                }
            }
        }
    }
    @res = uniqnumsort(@res);

    return intersect_aref(\@dates, \@res);
}

# DATERANGE_SPEC_WITHYEAR = RANGE(SINGLE_DATE_WITHYEAR)
package AdvPowerMgr::Plugin::Program::Date::DateRangeSpecWithYear;

use strict;
use warnings;

use base qw( AdvPowerMgr::Plugin::Program::Date::SingleDateWithYear );

use AdvPowerMgr::Date;
use AdvPowerMgr::Errors;
use AdvPowerMgr::I18N qw( i18n );
use AdvPowerMgr::Utils qw( intersect_aref range range_splitter );

sub _regex {
    my $class = shift;
    return range(AdvPowerMgr::Plugin::Program::Date::SingleDateWithYear->_regex());
}

sub text_normalizer {
    my $class = shift;
    my $value = param_defined(shift);

    my ($first, $last) = range_splitter(AdvPowerMgr::Plugin::Program::Date::SingleDateWithYear->_regex(), $value);
    my $r1 = AdvPowerMgr::Plugin::Program::Date::SingleDateWithYear->text_normalizer($first);
    my $r2 = AdvPowerMgr::Plugin::Program::Date::SingleDateWithYear->text_normalizer($last);

    return $r1 eq $r2 ? $r1 : "$r1-$r2";
}

sub text_trigger {
    my $self = shift;

    my ($first, $last) = range_splitter(AdvPowerMgr::Plugin::Program::Date::SingleDateWithYear->_regex(), $self->text());
    my @res = ();

    $first = AdvPowerMgr::Plugin::Program::Date::SingleDateWithYear->new($first)->_repr();
    $last = AdvPowerMgr::Plugin::Program::Date::SingleDateWithYear->new($last)->_repr();

    die i18n('program.invalid_range', $self->text()) if $last < $first;

    push(@res, $first);
    push(@res, $last) unless $last == $first;

    $self->_repr(\@res);
}

sub compute_ts {
    my $self = shift;
    my $when = AdvPowerMgr::Date->new(defined($_[0]) ? $_[0] : time()); shift;
    my $daysahead = shift; defined($daysahead) or $daysahead = 7;

    my ($first, $last) = @{$self->_repr()};

    unless (defined($last)) {
        return undef if $first < $when;
        return [] if $first > $when + $daysahead;
        return [ $first->ts() ];
    }

    return undef if $last < $when;

    my @res = ();
    my $date = $first > $when ? $first : $when;
    my $maxdate = $last > $when + $daysahead ? $when + $daysahead : $last;

    while ($date <= $maxdate) {
        push(@res, $date->ts());
        $date++;
    }

    return \@res;
}

# DATERANGE_SPEC := ( DATERANGE_SPEC_NOYEAR | DATERANGE_SPEC_WITHYEAR )
package AdvPowerMgr::Plugin::Program::Date::DateRangeSpec;

use strict;
use warnings;

use base qw( AdvPowerMgr::Plugin::Program::Date::Generic );

use AdvPowerMgr::Date;
use AdvPowerMgr::Errors;
use AdvPowerMgr::Utils qw( orregex );

our $regex_noyear = AdvPowerMgr::Plugin::Program::Date::DateRangeSpecNoYear->_regex();
our $regex_withyear = AdvPowerMgr::Plugin::Program::Date::DateRangeSpecWithYear->_regex();

sub _regex {
    my $class = shift;
    return orregex($regex_noyear, $regex_withyear);
}

sub text_normalizer {
    my $class = shift;
    my $value = param_defined(shift);

    if ($value =~ m#^$regex_noyear$#i) {
        return AdvPowerMgr::Plugin::Program::Date::DateRangeSpecNoYear->text_normalizer($value);
    } else {
        $value =~ m#^$regex_withyear$#i;
        return AdvPowerMgr::Plugin::Program::Date::DateRangeSpecWithYear->text_normalizer($value);
    }
}

sub text_trigger {
    my $self = shift;

    if ($self->text() =~ m#^$regex_noyear$#i) {
        $self->_repr(AdvPowerMgr::Plugin::Program::Date::DateRangeSpecNoYear->new($self->text()));
    } else {
        $self->text() =~ m#^$regex_withyear$#i;
        $self->_repr(AdvPowerMgr::Plugin::Program::Date::DateRangeSpecWithYear->new($self->text()));
    }
}

sub compute_ts {
    my $self = shift;
    my $when = AdvPowerMgr::Date->new(defined($_[0]) ? $_[0] : time()); shift;
    my $daysahead = shift; defined($daysahead) or $daysahead = 7;

    return $self->_repr()->compute_ts($when, $daysahead);
}

# MONTH_NUMERALWEEKDAY_SPEC := MONTH_SELECTOR [ YEAR_LRANGE ] NUMERALWEEKDAY_SELECTOR
package AdvPowerMgr::Plugin::Program::Date::MonthNumeralWeekdaySpec;

use strict;
use warnings;

use base qw( AdvPowerMgr::Plugin::Program::Date::Generic );

use AdvPowerMgr::Date;
use AdvPowerMgr::Errors;
use AdvPowerMgr::Utils qw( intersect_aref );

our $regex = AdvPowerMgr::Plugin::Program::Date::MonthSelector->_regex() . "(?:\\s+" . AdvPowerMgr::Plugin::Program::Date::YearLRange->_regex() . ")?\\s+" . AdvPowerMgr::Plugin::Program::Date::NumeralWeekdaySelector->_regex();
our $regex_matcher = "^(" . AdvPowerMgr::Plugin::Program::Date::MonthSelector->_regex() . ")(?:\\s+(" . AdvPowerMgr::Plugin::Program::Date::YearLRange->_regex() . "))?\\s+(" . AdvPowerMgr::Plugin::Program::Date::NumeralWeekdaySelector->_regex() . ")\$";

sub _regex {
    my $class = shift;
    return $regex;
}

sub text_normalizer {
    my $class = shift;
    my $value = param_defined(shift);

    my @res = ();
    $value =~ m#$regex_matcher#i;
    push(@res, AdvPowerMgr::Plugin::Program::Date::MonthSelector->text_normalizer($1));
    push(@res, AdvPowerMgr::Plugin::Program::Date::YearLRange->text_normalizer($2)) if defined($2);
    push(@res, AdvPowerMgr::Plugin::Program::Date::NumeralWeekdaySelector->text_normalizer($3));

    return join(" ", @res);
}

sub text_trigger {
    my $self = shift;

    $self->text() =~ m#$regex_matcher#i;

    my @res = ();

    push(@res, AdvPowerMgr::Plugin::Program::Date::MonthSelector->new($1));
    push(@res, AdvPowerMgr::Plugin::Program::Date::YearLRange->new($2)) if defined($2);
    push(@res, AdvPowerMgr::Plugin::Program::Date::NumeralWeekdaySelector->new($3));

    $self->_repr(\@res);
}

sub compute_ts {
    my $self = shift;
    my $when = AdvPowerMgr::Date->new(defined($_[0]) ? $_[0] : time()); shift;
    my $daysahead = shift; defined($daysahead) or $daysahead = 7;

    my @arr = ();
    foreach my $obj (@{$self->_repr()}) {
        push(@arr, $obj->compute_ts($when, $daysahead));
    }
    return intersect_aref(@arr);
}

# MONTH_WEEKDAY_SPEC := [ MONTH_SELECTOR ] [ YEAR_LRANGE ] WEEKDAY_SELECTOR
package AdvPowerMgr::Plugin::Program::Date::MonthWeekdaySpec;

use strict;
use warnings;

use base qw( AdvPowerMgr::Plugin::Program::Date::Generic );

use AdvPowerMgr::Date;
use AdvPowerMgr::Errors;
use AdvPowerMgr::Utils qw( intersect_aref );

our $regex = "(?:" . AdvPowerMgr::Plugin::Program::Date::MonthSelector->_regex() . "\\s+)?(?:" . AdvPowerMgr::Plugin::Program::Date::YearLRange->_regex() . "\\s+)?" . AdvPowerMgr::Plugin::Program::Date::WeekdaySelector->_regex();
our $regex_matcher = "^(?:(" . AdvPowerMgr::Plugin::Program::Date::MonthSelector->_regex() . ")\\s+)?(?:(" . AdvPowerMgr::Plugin::Program::Date::YearLRange->_regex() . ")\\s+)?(" . AdvPowerMgr::Plugin::Program::Date::WeekdaySelector->_regex() . ")\$";

sub _regex {
    my $class = shift;
    return $regex;
}

sub text_normalizer {
    my $class = shift;
    my $value = param_defined(shift);

    $value =~ m#$regex_matcher#i;
    my @res = ();
    push(@res, AdvPowerMgr::Plugin::Program::Date::MonthSelector->text_normalizer($1)) if defined($1);
    push(@res, AdvPowerMgr::Plugin::Program::Date::YearLRange->text_normalizer($2)) if defined($2);
    push(@res, AdvPowerMgr::Plugin::Program::Date::WeekdaySelector->text_normalizer($3));

    return join(" ", @res);
}

sub text_trigger {
    my $self = shift;

    $self->text() =~ m#$regex_matcher#i;

    my @res = ();

    push(@res, AdvPowerMgr::Plugin::Program::Date::MonthSelector->new($1)) if defined($1);
    push(@res, AdvPowerMgr::Plugin::Program::Date::YearLRange->new($2)) if defined($2);
    push(@res, AdvPowerMgr::Plugin::Program::Date::WeekdaySelector->new($3));

    $self->_repr(\@res);
}

sub compute_ts {
    my $self = shift;
    my $when = AdvPowerMgr::Date->new(defined($_[0]) ? $_[0] : time()); shift;
    my $daysahead = shift; defined($daysahead) or $daysahead = 7;

    my @arr = ();
    foreach my $obj (@{$self->_repr()}) {
        push(@arr, $obj->compute_ts($when, $daysahead));
    }
    return intersect_aref(@arr);
}

# YEAR_WEEK_SPEC := WEEK_SELECTOR [ YEAR_LRANGE ] WEEKDAY_SELECTOR
package AdvPowerMgr::Plugin::Program::Date::YearWeekSpec;

use strict;
use warnings;

use base qw( AdvPowerMgr::Plugin::Program::Date::Generic );

use AdvPowerMgr::Date;
use AdvPowerMgr::Errors;
use AdvPowerMgr::Utils qw( intersect_aref );

our $regex = AdvPowerMgr::Plugin::Program::Date::WeekSelector->_regex() . "\\s+(?:" . AdvPowerMgr::Plugin::Program::Date::YearLRange->_regex() . "\\s+)?" . AdvPowerMgr::Plugin::Program::Date::WeekdaySelector->_regex();
our $regex_matcher = "^(" . AdvPowerMgr::Plugin::Program::Date::WeekSelector->_regex() . ")\\s+(?:(" . AdvPowerMgr::Plugin::Program::Date::YearLRange->_regex() . ")\\s+)?(" . AdvPowerMgr::Plugin::Program::Date::WeekdaySelector->_regex() . ")\$";

sub _regex {
    my $class = shift;
    return $regex;
}

sub text_normalizer {
    my $class = shift;
    my $value = param_defined(shift);

    $value =~ m#$regex_matcher#i;
    my @res = ();
    push(@res, AdvPowerMgr::Plugin::Program::Date::WeekSelector->text_normalizer($1));
    push(@res, AdvPowerMgr::Plugin::Program::Date::YearLRange->text_normalizer($2)) if defined($2);
    push(@res, AdvPowerMgr::Plugin::Program::Date::WeekdaySelector->text_normalizer($3));

    return join(" ", @res);
}

sub text_trigger {
    my $self = shift;

    $self->text() =~ m#$regex_matcher#i;

    my @res = ();

    push(@res, AdvPowerMgr::Plugin::Program::Date::WeekSelector->new($1));
    push(@res, AdvPowerMgr::Plugin::Program::Date::YearLRange->new($2)) if defined($2);
    push(@res, AdvPowerMgr::Plugin::Program::Date::WeekdaySelector->new($3));

    $self->_repr(\@res);
}

sub compute_ts {
    my $self = shift;
    my $when = AdvPowerMgr::Date->new(defined($_[0]) ? $_[0] : time()); shift;
    my $daysahead = shift; defined($daysahead) or $daysahead = 7;

    my @arr = ();
    foreach my $obj (@{$self->_repr()}) {
        push(@arr, $obj->compute_ts($when, $daysahead));
    }
    return intersect_aref(@arr);
}

# YEARDAY_SPEC := yearday YEARDAY_LRANGE [ YEAR_LRANGE ]
package AdvPowerMgr::Plugin::Program::Date::YeardaySpec;

use strict;
use warnings;

use base qw( AdvPowerMgr::Plugin::Program::Date::Generic );

use AdvPowerMgr::Date;
use AdvPowerMgr::Errors;
use AdvPowerMgr::Utils qw( intersect_aref );

our $regex = "yearday\\s+" . AdvPowerMgr::Plugin::Program::Date::YeardayLRange->_regex() . "(?:\\s+" . AdvPowerMgr::Plugin::Program::Date::YearLRange->_regex() . ")?";
our $regex_matcher = "^yearday\\s+(" . AdvPowerMgr::Plugin::Program::Date::YeardayLRange->_regex() . ")(?:\\s+(" . AdvPowerMgr::Plugin::Program::Date::YearLRange->_regex() . "))?\$";

sub _regex {
    my $class = shift;
    return $regex;
}

sub text_normalizer {
    my $class = shift;
    my $value = param_defined(shift);

    my @res = ();
    $value =~ m#$regex_matcher#i;
    push(@res, "yearday");
    push(@res, AdvPowerMgr::Plugin::Program::Date::YeardayLRange->text_normalizer($1));
    push(@res, AdvPowerMgr::Plugin::Program::Date::YearLRange->text_normalizer($2)) if defined($2);

    return join(" ", @res);
}

sub text_trigger {
    my $self = shift;

    $self->text() =~ m#$regex_matcher#i;

    my @res = ();

    push(@res, AdvPowerMgr::Plugin::Program::Date::YeardayLRange->new($1));
    push(@res, AdvPowerMgr::Plugin::Program::Date::YearLRange->new($2)) if defined($2);

    $self->_repr(\@res);
}

sub compute_ts {
    my $self = shift;
    my $when = AdvPowerMgr::Date->new(defined($_[0]) ? $_[0] : time()); shift;
    my $daysahead = shift; defined($daysahead) or $daysahead = 7;

    my @arr = ();
    foreach my $obj (@{$self->_repr()}) {
        push(@arr, $obj->compute_ts($when, $daysahead));
    }
    return intersect_aref(@arr);
}

# SPEC := (
#   DATE_SPEC
#   DATERANGE_SPEC
#   MONTH_NUMERALWEEKDAY_SPEC
#   MONTH_WEEKDAY_SPEC
#   YEAR_WEEK_SPEC
#   YEARDAY_SPEC
# )
package AdvPowerMgr::Plugin::Program::Date::Spec;

use strict;
use warnings;

use base qw( AdvPowerMgr::Plugin::Program::Date::Generic );

use AdvPowerMgr::Date;
use AdvPowerMgr::Errors;
use AdvPowerMgr::Utils qw( intersect_aref orregex );

our $regex1 = AdvPowerMgr::Plugin::Program::Date::DateSpec->_regex();
our $regex2 = AdvPowerMgr::Plugin::Program::Date::DateRangeSpec->_regex();
our $regex3 = AdvPowerMgr::Plugin::Program::Date::MonthNumeralWeekdaySpec->_regex();
our $regex4 = AdvPowerMgr::Plugin::Program::Date::MonthWeekdaySpec->_regex();
our $regex5 = AdvPowerMgr::Plugin::Program::Date::YearWeekSpec->_regex();
our $regex6 = AdvPowerMgr::Plugin::Program::Date::YeardaySpec->_regex();

sub _regex {
    my $class = shift;
    return orregex($regex1, $regex2, $regex3, $regex4, $regex5, $regex6);
}

sub text_normalizer {
    my $class = shift;
    my $value = param_defined(shift);

    my @res = ();
    if ($value =~ m#^$regex1$#i) {
        return AdvPowerMgr::Plugin::Program::Date::DateSpec->text_normalizer($value);
    } elsif ($value =~ m#^$regex2$#i) {
        return AdvPowerMgr::Plugin::Program::Date::DateRangeSpec->text_normalizer($value);
    } elsif ($value =~ m#^$regex3$#i) {
        return AdvPowerMgr::Plugin::Program::Date::MonthNumeralWeekdaySpec->text_normalizer($value);
    } elsif ($value =~ m#^$regex4$#i) {
        return AdvPowerMgr::Plugin::Program::Date::MonthWeekdaySpec->text_normalizer($value);
    } elsif ($value =~ m#^$regex5$#i) {
        return AdvPowerMgr::Plugin::Program::Date::YearWeekSpec->text_normalizer($value);
    } else {
        return AdvPowerMgr::Plugin::Program::Date::YeardaySpec->text_normalizer($value);
    }
}

sub text_trigger {
    my $self = shift;

    if ($self->text() =~ m#^$regex1$#i) {
        $self->_repr(AdvPowerMgr::Plugin::Program::Date::DateSpec->new($self->text()));
    } elsif ($self->text() =~ m#^$regex2$#i) {
        $self->_repr(AdvPowerMgr::Plugin::Program::Date::DateRangeSpec->new($self->text()));
    } elsif ($self->text() =~ m#^$regex3$#i) {
        $self->_repr(AdvPowerMgr::Plugin::Program::Date::MonthNumeralWeekdaySpec->new($self->text()));
    } elsif ($self->text() =~ m#^$regex4$#i) {
        $self->_repr(AdvPowerMgr::Plugin::Program::Date::MonthWeekdaySpec->new($self->text()));
    } elsif ($self->text() =~ m#^$regex5$#i) {
        $self->_repr(AdvPowerMgr::Plugin::Program::Date::YearWeekSpec->new($self->text()));
    } else {
        $self->_repr(AdvPowerMgr::Plugin::Program::Date::YeardaySpec->new($self->text()));
    }
}

sub compute_ts {
    my $self = shift;
    my $when = AdvPowerMgr::Date->new(defined($_[0]) ? $_[0] : time()); shift;
    my $daysahead = shift; defined($daysahead) or $daysahead = 7;

    return $self->_repr()->compute_ts($when, $daysahead);
}

# DATE := LIST(SPEC)
package AdvPowerMgr::Plugin::Program::Date;

use strict;
use warnings;

use base qw( AdvPowerMgr::Plugin::Program::Date::Spec );

use AdvPowerMgr::Date;
use AdvPowerMgr::Errors;
use AdvPowerMgr::Utils qw( list list_splitter uniqnumsort );

sub _regex {
    my $class = shift;
    return list($class->SUPER::_regex());
}

sub text_normalizer {
    my $class = shift;
    my $value = param_defined(shift);

    my @res = ();
    my @list = list_splitter($class->SUPER::_regex(), $value);
    foreach my $item (@list) {
        push(@res, AdvPowerMgr::Plugin::Program::Date::Spec->text_normalizer($item));
    }
    return join(", ", @res);
}

sub text_trigger {
    my $self = shift;

    my @list = list_splitter($self->SUPER::_regex(), $self->text());
    my @res = ();

    foreach my $item (@list) {
        push(@res, AdvPowerMgr::Plugin::Program::Date::Spec->new($item));
    }

    $self->_repr(\@res);
}

sub compute_ts {
    my $self = shift;
    my $when = AdvPowerMgr::Date->new(defined($_[0]) ? $_[0] : time()); shift;
    my $daysahead = shift; defined($daysahead) or $daysahead = 7;

    my $res = undef;

    foreach my $item (@{$self->_repr()}) {
        my $r = $item->compute_ts($when, $daysahead);
        if (defined($r)) {
            $res = [] unless defined($res);
            push(@{$res}, @{$r});
        }
    }

    if (defined($res)) {
        return [ uniqnumsort(@$res) ];
    } else {
        return undef;
    }
}

1;

# vi:set autoindent tabstop=8 shiftwidth=4 softtabstop=4 expandtab:
