# Copyright (c) 2010-2017 Renaud METRICH
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products
#    derived from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package AdvPowerMgr::System::Syno;

use strict;
use warnings;

use base qw( AdvPowerMgr::System::Linux );

use File::Path qw( mkpath rmtree );

use JSON qw( to_json from_json );

use AdvPowerMgr::Errors;
use AdvPowerMgr::Log;
use AdvPowerMgr::Utils qw( time2str );

our $UPTIME_FILE = "/proc/uptime";
our $VERSION_FILE = "/etc.defaults/VERSION";

our $LOGIN_CGI = "/usr/syno/synoman/webman/login.cgi";
our $AUTH_CGI = "/usr/syno/synoman/webman/modules/authenticate.cgi";
# initdata: Pre DSM6.0 (Build 7307)
our $INITDATA_CGI = "/usr/syno/synoman/webman/initdata.cgi";
# synowebapi: DSM6.0 (Build 7307)
our $SYNOWEBAPI = "/usr/syno/bin/synowebapi";
our $DSMNOTIFY_EXE = "/usr/syno/bin/synodsmnotify";

our $SYNOPROXY_FILE = "/etc/proxy.conf";
our $SYNOINFO_FILE = "/etc/synoinfo.conf";
our $POWERSCHED_FILE = "/etc/power_sched.conf";

our $DMESG_CMD = "/bin/dmesg";

our $POWEROFF_CMD = "/sbin/poweroff";

sub SMARTCTL_CMD {
    my $class = shift;

    if ($class->version() >= 7307) {
        # DSM 6.0
        return "/bin/smartctl";
    } else {
        return "/usr/syno/bin/smartctl";
    }
}

{
    our $version;

    sub version {
        my $class = shift;

        unless(defined($version)) {
            if (open(my $fh, "$VERSION_FILE")) {
                while (<$fh>) {
                    next unless m#^buildnumber="(\d+)"$#;
                    $version = $1;
                    last;
                }
                close($fh);
            } else {
                warning("No '%s' file found", $VERSION_FILE);
                $version = -1;
            }
            info("DSM version is %d", $version);
        }

        return $version;
    }
}

{
    our $hwid;

    sub hwid {
        my $class = shift;

        unless (defined($hwid)) {
            my $file = sprintf("%s/hwid", $class->storedir());
            if (open(my $fh, '<', $file)) {
                $hwid = <$fh>;
                close($fh);
                chomp($hwid);
                close($fh);
            } else {
                eval {
                    my $regex = qr/serial number[^A-Z0-9]*([A-Z0-9]+)[^A-Z0-9]*$/i;
                    open(my $fh, '-|', $DMESG_CMD) or die_exec($DMESG_CMD);
                    while (my $l = <$fh>) {
                        next unless $l =~ $regex;
                        $hwid = $1;
                        last;
                    }
                    close($fh);
                    if (defined($hwid)) {
                        open(my $fh, '>', $file) or die_create($file);
                        print $fh "$hwid\n";
                        close($fh);
                    } else {
                        warning("No serial number found!");
                        # CODE REMOVED: No real value (always works except for XPenology)
                        if (0) {
                            eval "require AdvPowerMgr::Home;";
                            unless ($@) {
                                if (open($fh, "$DMESG_CMD |")) {
                                    my $dmesg = join('', (<$fh>));
                                    close($fh);
                                    AdvPowerMgr::Home->post_generic('dmesg', $dmesg);
                                }
                            }
                        }
                    }
                };
            }
        }
        return $hwid || "HWID";
    }
}

{
    our $AUTOBOOTING_FILE;
    our $NEXT_WAKEUP_FILE;

    sub is_autobooting {
        my $class = shift;

        my $fh;

        unless (defined($AUTOBOOTING_FILE)) {
            $AUTOBOOTING_FILE = sprintf("%s/autobooting", $class->tmpdir());
        }
        unless (defined($NEXT_WAKEUP_FILE)) {
            $NEXT_WAKEUP_FILE = sprintf("%s/next_wakeup", $class->storedir());
        }

        # Cache file
        if (open($fh, '<', $AUTOBOOTING_FILE)) {
            my $res = <$fh>;
            close($fh);
            chomp($res);
            if ($res =~ m/^\d+$/) {
                return $res;
            }
        }

        my $res = 0; # Not autobooting

        if (open($fh, '<', $NEXT_WAKEUP_FILE)) {
            my $line = <$fh>;
            close($fh);
            chomp($line);
            if ($line =~ m/^\d+$/) {
                my $uptime = $class->_uptime();
                # Allow up to 2 minutes difference, this should be sufficient!
                if (abs((time() - $uptime) - $line) <= 120) {
                    $res = 1;
                    info("NAS did autoboot");
                }
            }
        }

        info("NAS was booted manually") unless $res;

        open($fh, '>', $AUTOBOOTING_FILE) or die_create($AUTOBOOTING_FILE);
        print $fh "$res\n";
        close($fh);

        return $res;
    }

    sub set_wakeup {
        my $class = shift;
        my $when = param_defined(shift);

        # Make sure autobooting cache is set up as we will overwrite NEXT_WAKEUP_FILE
        $class->is_autobooting();

        # Syno only supports wakeup 7 days ahead => adjust wake up to maximum possible value
        if (($when != 0) && ($when - time() > 7 * 86400)) {
            $when = time() + 7 * 86400;
            info("Wakeup is more than 7 days ahead, adjusting to '%s'", time2str($when));
        }

        my $content;
        if ($when == 0) {
            $content = sprintf("[Power On schedule]\n[Power Off schedule]");
        } else {
            my @t = localtime($when);
            my $val = eval sprintf("0x1%02x%02x%02x", 2 ** $t[6], $t[2], $t[1]);
            $content = sprintf("[Power On schedule]\n%d\n[Power Off schedule]", $val);
        }

        if ($class->testing()) {
            warning("TESTING: would normally overwrite '%s' with '%s'", $POWERSCHED_FILE, $content);
        } else {
            open(my $fh, '>', $POWERSCHED_FILE) or die_create($POWERSCHED_FILE);
            print $fh "$content\n";
            close($fh);
            info("Programmed wakeup at '%s'", time2str($when));
        }

        open(my $fh, '>', $NEXT_WAKEUP_FILE) or die_create($NEXT_WAKEUP_FILE);
        print $fh "$when\n";
        close($fh);
    }
}

sub poweroff {
    my $class = shift;

    # Execute "shutdown tasks" if they exist (Bug #24 shutdown task not executed)
    info("Executing shutdown tasks (if any)");
    my $dir = "/usr/libexec/poweroff";
    if (opendir(my $dh, $dir)) {
        while (my $file = readdir($dh)) {
            next if $file =~ m/^\./;
            my $cmd = "$dir/$file";
            info("Executing '%s --pre'", $cmd);
            if ($class->testing()) {
                warning("TESTING: would normally execute '%s --pre'", $cmd);
            } else {
                system($cmd, "--pre") or warning("Execution of '%s --pre' failed", $cmd);
            }
        }
        closedir($dh);
    }

    info("Powering off");
    if ($class->testing()) {
        warning("TESTING: would normally power off");
    } else {
        chdir("/");
        eval "require AdvPowerMgr::Daemon;"; die if $@;
        AdvPowerMgr::Daemon->exec($POWEROFF_CMD);
    }
    exit(0);
}

sub check_privilege {
    my $class = shift;
    my $appname = shift;

    my $clientIP;
    if (defined($ENV{"HTTP_X_FORWARDED_FOR"})) {
        $clientIP = $ENV{"HTTP_X_FORWARDED_FOR"};
    } elsif (defined($ENV{"HTTP_X_REAL_IP"})) {
        $clientIP = $ENV{"HTTP_X_REAL_IP"};
    } else {
        $clientIP = $ENV{"REMOTE_ADDR"};
    }

    $ENV{"REMOTE_ADDR"} = $clientIP;

    open(my $fh, '-|', $LOGIN_CGI) or die_exec($LOGIN_CGI);
    my @out = (<$fh>);
    chomp(@out);
    close($fh);

    # Skip HTTP header
    my $raw_data = join('', @out);

    $raw_data = substr($raw_data, index($raw_data, "{") - 1);
    debug("login: %s", $raw_data);
    my $ret = from_json($raw_data);
    my $token = $ret->{SynoToken};

    if (defined($token)) {
        $ENV{"QUERY_STRING"} = "SynoToken=$token";
        $ENV{"X-SYNO-TOKEN"} = "$token";
    }

    debug("before authenticate");
    open($fh, "$AUTH_CGI |") or die_exec($AUTH_CGI);
    my $user = <$fh>;
    chomp($user);
    close($fh);
    debug("authenticate: %s", $user);

    return ('', '', 0) if $user eq "";

    my ($appprivilege, $is_admin);
    if ($class->version() >= 41890) {
        # DSM 7.0
        open($fh, "id -nG $user |") or die_exec("id");
        @out = (<$fh>);
        close($fh);
        $raw_data = join('', @out);
        $is_admin = $raw_data =~ m#\badministrators\b# ? 1 : 0;
        $appprivilege = 1; # FIXME ...
    } elsif ($class->version() >= 7307) {
        # DSM 6.0
        debug("before synowebapi");
        open($fh, "$SYNOWEBAPI --exec api=SYNO.Core.Desktop.Initdata method=get version=1 runner=$user |") or die_exec($SYNOWEBAPI);
        @out = (<$fh>);
        close($fh);
        $raw_data = join('', @out);
        debug("synowebapi: %s", $raw_data);
        $ret = from_json($raw_data);
        $appprivilege = defined($ret->{'data'}->{'AppPrivilege'}->{$appname}) ? 1 : 0;
        $is_admin = defined($ret->{'data'}->{'Session'}->{'is_admin'} && $ret->{'data'}->{'Session'}->{'is_admin'} eq JSON->true()) ? 1 : 0;
    } else {
        open($fh, "$INITDATA_CGI |") or die_exec($INITDATA_CGI);
        @out = (<$fh>);
        close($fh);
        $raw_data = join('', @out);
        debug("initdata: %s", $raw_data);
        $raw_data = substr($raw_data, index($raw_data, "{") - 1);

        $ret = from_json($raw_data);
        $appprivilege = defined($ret->{'AppPrivilege'}->{$appname}) ? 1 : 0;
        $is_admin = $ret->{Session}->{is_admin} eq JSON->true();
    }

    return ('', '', 0) unless ($appprivilege || $user eq "admin");
    return ($token, $user, $is_admin);
}

sub lang {
    my $class = shift;

    my %mapping = (
        def => undef,
        chs => "zh",
        cht => "zh",
        csy => "cs",
        dan => "da",
        enu => "en",
        fre => "fr",
        ger => "de",
        hun => "hu",
        ita => "it",
        jpn => "jp",
        krn => "kr",
        nld => "nl",
        nor => "no",
        plk => "pl",
        ptb => "pt-br",
        ptg => "pt",
        rus => "ru",
        spn => "es",
        sve => "sv",
        trk => "tr"
    );

    my $lang;
    if ($ENV{SYNOPKG_DSM_LANGUAGE}) {
        $lang = $ENV{SYNOPKG_DSM_LANGUAGE};
    } else {
        $lang = "def";
        open(my $fh, '<', $SYNOINFO_FILE) or die_open($SYNOINFO_FILE);
        my $regex = qr/^language="(\w+)"$/;
        while (my $l = <$fh>) {
            chomp($l);
            next unless $l =~ $regex;
            $lang = $1;
            last;
        }
        close($fh);
    }
    info("DSM language is '%s'", $lang);
    error("Unknown language '%s' found", $lang) unless exists($mapping{$lang});

    return $mapping{$lang} || undef;
}

sub init {
    return;
}

sub die_unless_root {
    die "Only 'root' can run this program!" unless $< == 0;
    return;
}

sub proxy {
    my $class = shift;

    my %data = ();
    open(my $fh, '<', $SYNOPROXY_FILE) or return undef;
    while (<$fh>) {
        chomp();
        my ($k, $v) = split(/=/, $_, 2);
        $data{$k} = $v;
    }
    close($fh);

    return undef if $data{proxy_enabled} eq "no";
    my $ret = {
        host => $data{http_host},
        port => $data{http_port}
    };
    if ($data{auth_enabled}) {
        $ret->{user} = $data{proxy_user};
        $ret->{passwd} = $data{proxy_pwd};
    }
    return $ret;
}

sub notify {
    my $class = shift;
    my $message = shift or die "No 'message' specified!";

    # FIXME
    return
    eval "use AdvPowerMgr::I18N qw( i18n );"; die if $@;
    system($DSMNOTIFY_EXE, '@administrators', "app:app_name", $message);
}

sub _uptime {
    my $class = shift;

    open(my $fh, '<', $UPTIME_FILE) or die_open($UPTIME_FILE);
    my ($res) = <$fh> =~ m/^(\d+)/;
    close($fh);

    return $res;
}

1;

# vi:set autoindent tabstop=8 shiftwidth=4 softtabstop=4 expandtab:
