# Copyright (c) 2010-2017 Renaud METRICH
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products
#    derived from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package AdvPowerMgr::Utils;

use strict;
use warnings;

use base qw( Exporter );

use Time::Local qw( timelocal );

use AdvPowerMgr::Errors;

our @EXPORT_OK;
push(@EXPORT_OK, qw( min_ts max_ts time2str str2time ipv4_to_hex uniq uniqnumsort millisleep txt2html ));
push(@EXPORT_OK, qw( intersect_aref remove_aref ));
push(@EXPORT_OK, qw( orregex range list lrange list_splitter range_splitter ));

sub txt2html {
    my $class = shift if eval { $_[0]->isa(__PACKAGE__) };
    my $txt = shift;

    s#\n#<br/>#g for $txt;
    return $txt;
}

sub millisleep {
    my $class = shift if eval { $_[0]->isa(__PACKAGE__) };
    my $value = shift;

    select(undef, undef, undef, $value);
}

sub max_ts {
    my $class = shift if eval { $_[0]->isa(__PACKAGE__) };

    my @res = sort { $a <=> $b } @_;

    return ($res[0] == -1 ? -1 : $res[@res - 1]);
}

sub min_ts {
    my $class = shift if eval { $_[0]->isa(__PACKAGE__) };

    my @res = sort { $a <=> $b } @_;

    return -1 if $res[0] == -1;

    foreach (@res) {
        return $_ if $_ > 0;
    }

    return 0;
}

# Compute the intersection of arrays (specified as array references)
# return an array reference
# or undef if some array is not defined
sub intersect_aref {
    my $class = shift if eval { $_[0]->isa(__PACKAGE__) };
    my @arr = @_;

    my %isect = ();

    my $items = 0;
    foreach my $aref (@arr) {
        return undef if !defined($aref);
        $items += scalar(@$aref);
    }
    return [] unless $items;

    foreach my $aref (@arr) {
        foreach my $e (@{$aref}) {
            $isect{$e}++;
        }
    }
    my $arr_size = scalar(@arr);

    my @res = ();
    foreach my $e (keys(%isect)) {
        push(@res, $e) if $isect{$e} == $arr_size;
    }
    return [ sort { $a <=> $b } @res ];
}

# Returns the values in first sorted array not in second sorted array (specified as array references)
sub remove_aref {
    my $class = shift if eval { $_[0]->isa(__PACKAGE__) };
    my $first = shift;
    my $second = shift;

    return $first unless defined($first) && defined($second);

    my @res = ( @{$first} );
    my $lastindex = 0;

    foreach my $s (@{$second}) {
        while ($lastindex < @res) {
            my $f = $res[$lastindex];
            last if $f > $s;
            if ($f == $s) {
                splice(@res, $lastindex, 1);
            } else {
                $lastindex++;
            }
        }
    }

    return \@res;
}

sub time2str {
    my $class = shift if eval { $_[0]->isa(__PACKAGE__) };
    my $ts = shift;

    return "None" if $ts == 0;
    return "Forever" if $ts == -1;

    my ($s, $m, $h, $dm, $mo, $y, $dw) = localtime($ts);
    $mo++; $y += 1900;

    return sprintf("%04d-%02d-%02d %02d:%02d:%02d", $y, $mo, $dm, $h, $m, $s);
}

{
    our $str2time_regex = qr/^(\d\d\d\d)-(\d\d)-(\d\d) (\d\d):(\d\d)(?::(\d\d))?$/;

    sub str2time {
        my $class = shift if eval { $_[0]->isa(__PACKAGE__) };
        my $date = param_defined(shift);

        $date =~ $str2time_regex or die_ie("Invalid date format '$date'");

        return timelocal($6 || 0, $5, $4, $3, $2 - 1, $1 - 1900);
    }
}

sub ipv4_to_hex {
    my $class = shift if eval { $_[0]->isa(__PACKAGE__) };
    my ($ip) = @_;
    return hex(sprintf("0x%02x%02x%02x%02x", split(/\./, $ip)));
}

sub uniqnumsort {
    my $class = shift if eval { $_[0]->isa(__PACKAGE__) };

    return uniq(sort { $a <=> $b } @_);
}

sub uniq {
    return () unless @_;
    my $class = shift if eval { $_[0]->isa(__PACKAGE__) };

    my $prev = 'nonesuch';
    return grep { $_ ne $prev && (($prev) = $_) } @_;
}

sub orregex {
    my $class = shift if eval { $_[0]->isa(__PACKAGE__) };
    return "(?:" . join('|', @_) . ")";
}

# LIST(X) := X [ , X ] ...
sub list {
    my $class = shift if eval { $_[0]->isa(__PACKAGE__) };
    my ($expr, $matcher) = @_;
    if (defined($matcher) && $matcher) {
        return "($expr)(?:\\s*,\\s*($expr(?:\\s*,\\s*$expr)*))?"; # $1 == item, $2 == rest or undef
    } else {
        return "(?:$expr)(?:\\s*,\\s*$expr)*";
    }
}

# RANGE(X) := X [ - X ]
sub range {
    my $class = shift if eval { $_[0]->isa(__PACKAGE__) };
    my ($expr, $matcher) = @_;
    if (defined($matcher) && $matcher) {
        return "($expr)(?:\\s*-\\s*($expr))?"; # $1 == left, $2 == right or undef
    } else {
        return "(?:$expr)(?:\\s*-\\s*$expr)?";
    }
}

# LRANGE(X) := LIST(RANGE(X))
sub lrange {
    my $class = shift if eval { $_[0]->isa(__PACKAGE__) };
    my ($expr, $matcher) = @_;
    # Use $matcher only for enclosing list, not range
    return list(range($expr), $matcher);
}

# Split a range or single value in range
# returns an array: ( first, last )
sub range_splitter {
    my $class = shift if eval { $_[0]->isa(__PACKAGE__) };
    my $baseregex = shift;
    my $text = shift;

    my $rangeregex = "^" . range($baseregex, 1) . "\$";

    my @res = ();

    $text =~ m#$rangeregex#i or die_ie("Text '$text': not a range");
    push(@res, $1);
    push(@res, defined($2) ? $2 : $1);

    return @res;
}

# Split a list
# returns an array of items
sub list_splitter {
    my $class = shift if eval { $_[0]->isa(__PACKAGE__) };
    my $baseregex = param_defined(shift);
    my $text = param_defined(shift);

    my $listregex = "^" . list($baseregex, 1) . "\$";
    my ($item, $rest) = (undef, $text);

    my @res = ();

    while ($rest =~ m#$listregex#i) {
        ($item, $rest) = ($1, defined($2) ? $2 : "");
        push(@res, $item);
    }
    die_ie("Text '$text': remaining string '$rest'") unless $rest eq "";

    return @res;
}

1;

# vi:set autoindent tabstop=8 shiftwidth=4 softtabstop=4 expandtab:
